from collections.abc import Iterator
from contextlib import contextmanager
from crystal.model import Project, Resource, ResourceGroup, RootResource
from crystal.tests.util.asserts import assertEqual
from crystal.tests.util.controls import click_button, click_checkbox, TreeItem
from crystal.tests.util.server import MockHttpServer, served_project
from crystal.tests.util.ssd import database_on_ssd
from crystal.tests.util.tasks import wait_for_download_to_start_and_finish
from crystal.tests.util.wait import (
    first_child_of_tree_item_is_not_loading_condition,
    tree_has_no_children_condition, tree_item_has_no_children_condition,
    wait_for,
)
from crystal.tests.util.windows import (
    EntityTree, MainWindow, NewGroupDialog, NewRootUrlDialog,
    OpenOrCreateDialog,
)
from crystal.tests.util.xurlparse import urlpatternparse
from crystal.util.wx_dialog import mocked_show_modal
from crystal.util.xos import is_windows
import re
from unittest import skip
from unittest.mock import patch
import wx

# === Test: Create & Delete Standalone ===

async def test_can_create_group_with_source(
        *, with_source: bool=True,
        add_surrounding_whitespace: bool=False) -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        comic_pattern = sp.get_request_url('https://xkcd.com/#/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, _):
            root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
            () = root_ti.Children
            
            # If will need a source later, create one now
            if with_source:
                assert mw.new_root_url_button.Enabled
                click_button(mw.new_root_url_button)
                nud = await NewRootUrlDialog.wait_for()
                
                nud.name_field.Value = 'Home'
                nud.url_field.Value = home_url
                nud.do_not_download_immediately()
                nud.do_not_set_default_url_prefix()
                await nud.ok()
                
                home_ti = root_ti.find_child(home_url)
            
            # Create a group
            if True:
                selected_ti = TreeItem.GetSelection(mw.entity_tree.window)
                if with_source:
                    # Ensure source selected
                    assert selected_ti == home_ti
                else:
                    # Ensure nothing selected
                    assert (selected_ti is None) or (selected_ti == root_ti)
                
                assert mw.new_group_button.Enabled
                click_button(mw.new_group_button)
                ngd = await NewGroupDialog.wait_for()
                
                # Ensure prepopulates reasonable information
                if with_source:
                    assert home_url == ngd.pattern_field.Value
                    assert 'Home' == ngd.name_field.Value
                    assert None == ngd.source
                else:
                    assert '' == ngd.pattern_field.Value
                    assert '' == ngd.name_field.Value
                    assert None == ngd.source
                assert ngd.pattern_field.HasFocus  # default focused field
                
                # Input new URL pattern with wildcard, to match comics
                ngd.pattern_field.Value = (
                    (' ' if add_surrounding_whitespace else '') +
                    comic_pattern +
                    (' ' if add_surrounding_whitespace else '')
                )
                
                # Ensure preview members show the new matching URLs (i.e. none)
                member_urls = [
                    ngd.preview_members_list.GetString(i)
                    for i in range(ngd.preview_members_list.GetCount())
                ]
                assert [] == member_urls  # no comics discovered yet
                
                if with_source:
                    ngd.source = 'Home'
                ngd.name_field.Value = 'Comic'
                await ngd.ok()
                
                # Ensure appearance is correct
                comic_ti = root_ti.find_child(comic_pattern)
                assert f'{comic_pattern} - Comic' == comic_ti.Text
                await _assert_tree_item_icon_tooltip_contains(comic_ti, 'Group')
                
                if not with_source:
                    # Ensure new group is selected automatically,
                    # given that nothing was previously selected
                    selected_ti = TreeItem.GetSelection(mw.entity_tree.window)
                    assert selected_ti == comic_ti
            
            # Forget group
            if True:
                comic_ti.SelectItem()
                assert mw.forget_button.IsEnabled()
                click_button(mw.forget_button)
                
                # Ensure cannot find group
                assert None == root_ti.try_find_child(comic_pattern)
                selected_ti = TreeItem.GetSelection(mw.entity_tree.window)
                if with_source and is_windows():
                    # Windows will retarget the selection to the remaining node at the root
                    assert selected_ti == home_ti
                else:
                    # Other platforms will retarget the selection to nothing
                    assert (selected_ti is None) or (selected_ti == root_ti)


async def test_can_create_group_with_no_source() -> None:
    await test_can_create_group_with_source(with_source=False)


async def test_cannot_create_group_with_empty_url_pattern() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, _):
            assert mw.new_group_button.Enabled
            click_button(mw.new_group_button)
            ngd = await NewGroupDialog.wait_for()
            
            ngd.pattern_field.Value = ''
            ngd.name_field.Value = 'Comic'
            with patch(
                    'crystal.browser.new_group.ShowModal',
                    mocked_show_modal('cr-empty-url-pattern', wx.ID_OK)
                    ) as show_modal_method:
                click_button(ngd.ok_button)
                assert 1 == show_modal_method.call_count
            
            await ngd.cancel()


@skip('covered by: test_can_create_group_with_source')
async def test_can_forget_group() -> None:
    pass


@skip('not yet automated')
async def test_when_forget_group_then_related_root_urls_and_revisions_are_not_deleted() -> None:
    pass


async def test_given_url_pattern_with_surrounding_whitespace_when_create_group_then_surrounding_whitespace_ignored() -> None:
    await test_can_create_group_with_source(add_surrounding_whitespace=True)


# === Test: Create & Delete from Links ===

async def test_given_resource_node_with_multiple_link_children_matching_url_pattern_can_create_new_group_to_bundle_those_links_together() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        if True:
            home_url = sp.get_request_url('https://xkcd.com/')
            
            comic1_url = sp.get_request_url('https://xkcd.com/1/')
            comic2_url = sp.get_request_url('https://xkcd.com/2/')
            comic_pattern = sp.get_request_url('https://xkcd.com/#/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            # Create home URL
            if True:
                root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                () = root_ti.Children
                
                assert mw.new_root_url_button.Enabled
                click_button(mw.new_root_url_button)
                nud = await NewRootUrlDialog.wait_for()
                
                nud.name_field.Value = 'Home'
                nud.url_field.Value = home_url
                nud.do_not_download_immediately()
                await nud.ok()
                (home_ti,) = root_ti.Children
            
            # Expand home URL
            home_ti.Expand()
            await wait_for_download_to_start_and_finish(mw.task_tree)
            assert first_child_of_tree_item_is_not_loading_condition(home_ti)()
            
            # Select a comic link from the home URL
            comic1_ti = home_ti.find_child(comic1_url, project.default_url_prefix)  # ensure did find sub-resource for Comic #1
            assert f'{urlpatternparse(comic1_url).path} - Link: |<, Link: |<' == comic1_ti.Text  # ensure expected links grouped
            comic1_ti.SelectItem()
            
            # Create a group to bundle the comic links together
            if True:
                assert mw.new_group_button.Enabled
                click_button(mw.new_group_button)
                ngd = await NewGroupDialog.wait_for()
                
                # Ensure prepopulates reasonable information
                assert comic1_url == ngd.pattern_field.Value  # default pattern = (from resource)
                assert '|<' == ngd.name_field.Value  # default name = (from first text link)
                assert 'Home' == ngd.source  # default source = (from resource parent)
                assert ngd.pattern_field.HasFocus  # default focused field
                
                # Ensure preview members show the 1 URL
                assert (
                    ngd.preview_members_pane is None or  # always expanded
                    ngd.preview_members_pane.IsExpanded()  # expanded by default
                )
                member_urls = [
                    ngd.preview_members_list.GetString(i)
                    for i in range(ngd.preview_members_list.GetCount())
                ]
                assert [comic1_url] == member_urls  # contains exact match of pattern
                
                # Input new URL pattern with wildcard, to match other comics
                ngd.pattern_field.Value = comic_pattern
                
                # Ensure preview members show the new matching URLs
                member_urls = [
                    ngd.preview_members_list.GetString(i)
                    for i in range(ngd.preview_members_list.GetCount())
                ]
                assert comic1_url in member_urls  # contains first comic
                assert len(member_urls) >= 2  # contains last comic too
                
                # Input new name
                ngd.name_field.Value = 'Comic'
                
                await ngd.ok()
            
            # 1. Ensure the new resource group does now bundle the comic links together
            # 2. Ensure the bundled link is selected immediately after closing the add group dialog
            if True:
                grouped_subresources_ti = home_ti.find_child(comic_pattern, project.default_url_prefix)  # ensure did find grouped sub-resources
                assert re.fullmatch(
                    rf'{re.escape(urlpatternparse(comic_pattern).path)} - \d+ of Comic',  # title format of grouped sub-resources
                    grouped_subresources_ti.Text)
                
                # Ensure the bundled link is selected immediately after closing the add group dialog
                assert grouped_subresources_ti.IsExpanded()
                comic1_ti = grouped_subresources_ti.find_child(comic1_url, project.default_url_prefix)  # contains first comic
                assert len(grouped_subresources_ti.Children) >= 2  # contains last comic too
                assert comic1_ti.IsSelected()
            
            # Download the group to download the links
            if True:
                grouped_subresources_ti.SelectItem()
                assert mw.download_button.IsEnabled()
                await mw.click_download_button()
                await wait_for_download_to_start_and_finish(mw.task_tree)
            
            # Forget the group to unbundle the links
            if True:
                grouped_subresources_ti.SelectItem()
                assert mw.forget_button.IsEnabled()
                click_button(mw.forget_button)
                
                # 1. Ensure can find the first unbundled link
                # 2. Ensure that first unbundled link is selected immediately after forgetting the group
                comic1_ti = home_ti.find_child(comic1_url, project.default_url_prefix)  # ensure did find sub-resource for Comic #1
                assert comic1_ti.IsSelected()


@skip('covered by: test_given_resource_node_with_multiple_link_children_matching_url_pattern_can_create_new_group_to_bundle_those_links_together')
async def test_given_resource_node_with_multiple_link_children_bundled_as_a_group_can_easily_download_the_group_to_download_the_links() -> None:
    pass


@skip('covered by: test_given_resource_node_with_multiple_link_children_matching_url_pattern_can_create_new_group_to_bundle_those_links_together')
async def test_given_resource_node_with_multiple_link_children_bundled_as_a_group_can_easily_forget_the_group_to_unbundle_the_links() -> None:
    pass


# === Test: Suggested Source when Create from Selection ===

async def test_given_node_is_selected_in_entity_tree_when_press_new_group_button_then_dialog_appears_with_reasonable_suggested_source() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        if True:
            home_url = sp.get_request_url('https://xkcd.com/')
            about_url = sp.get_request_url('https://xkcd.com/about')
            archive_url = sp.get_request_url('https://xkcd.com/archive')
            twitter_url = sp.get_request_url('https://twitter.com/xkcd/')
            stylesheet_url = sp.get_request_url('https://xkcd.com/s/7d94e0.css')
            
            comic1_url = sp.get_request_url('https://xkcd.com/1/')
            comic2_url = sp.get_request_url('https://xkcd.com/2/')
            comic_pattern = sp.get_request_url('https://xkcd.com/#/')
            
            atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
            rss_feed_url = sp.get_request_url('https://xkcd.com/rss.xml')
            feed_pattern = sp.get_request_url('https://xkcd.com/*.xml')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            # Create entities, which will appear in Entity Tree
            if True:
                home_rr = RootResource(project, 'Home', Resource(project, home_url))
                RootResource(project, 'About', Resource(project, about_url))
                RootResource(project, 'Atom Feed', Resource(project, atom_feed_url))
                
                ResourceGroup(project, 'Feed', feed_pattern, source=None)
                ResourceGroup(project, 'Comic', comic_pattern, source=home_rr)
            
            root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
            
            # Set Default URL Prefix so that Offsite clusters appear
            await mw.entity_tree.set_default_directory_to_entity_at_tree_item(
                _find_child(root_ti, home_url))
            url_prefix = project.default_url_prefix
            
            # Locate related entity tree nodes
            if True:
                # Children of RootNode
                home_rrn = _find_child(root_ti, home_url, url_prefix)
                feed_rgn = _find_child(root_ti, feed_pattern, url_prefix)
                comic_rgn = _find_child(root_ti, comic_pattern, url_prefix)
                
                # Children of RootResourceNode (a kind of _ResourceNode) that is inside RootNode
                await _expand_node(home_rrn, mw, will_download=True)
                home_rrn__atom_feed_rrn = _find_child(
                    home_rrn, atom_feed_url, url_prefix)
                home_rrn__about_rrn = _find_child(
                    home_rrn, about_url, url_prefix)
                home_rrn__feed_glrn = _find_child(
                    home_rrn, feed_pattern, url_prefix)
                await _expand_node(home_rrn__feed_glrn)
                home_rrn__feed_glrn__atom_feed_rrn = _find_child(
                    home_rrn__feed_glrn, atom_feed_url, url_prefix)
                home_rrn__feed_glrn__rss_feed_lrn = _find_child(
                    home_rrn__feed_glrn, rss_feed_url, url_prefix)
                home_rrn__archive_lrn = _find_child(
                    home_rrn, archive_url, url_prefix)
                home_rrn__offsite_cn = _find_child_by_title(
                    home_rrn, 'Offsite')
                await _expand_node(home_rrn__offsite_cn)
                home_rrn__offsite_cn__twitter_lrn = _find_child(
                    home_rrn__offsite_cn, twitter_url, url_prefix)
                home_rrn__embedded_cn = _find_child_by_title(
                    home_rrn, 'Embedded')
                await _expand_node(home_rrn__embedded_cn)
                home_rrn__embedded_cn__stylesheet_lrn = _find_child(
                    home_rrn__embedded_cn, stylesheet_url, url_prefix)
                
                # Children of ResourceGroupNode
                await _expand_node(feed_rgn)
                feed_rgn__atom_feed_rrn = _find_child(
                    feed_rgn, atom_feed_url, url_prefix)
                feed_rgn__rss_feed_nrn = _find_child(
                    feed_rgn, rss_feed_url, url_prefix)
                
                # ---
                
                # Children of LinkedResourceNode (a kind of _ResourceNode) that is inside GroupedLinkedResourceNode
                await _expand_node(home_rrn__feed_glrn__rss_feed_lrn, mw, will_download=True)
                home_rrn__feed_glrn__rss_feed_lrn__home_rrn = _find_child(
                    home_rrn__feed_glrn__rss_feed_lrn, home_url, url_prefix)
                
                # Children of LinkedResourceNode (a kind of _ResourceNode) that is inside RootResourceNode
                # HACK: Relies on Crystal's "Not in Archive" page to provide a
                #       link to the original URL (which is offsite)
                await _expand_node(home_rrn__archive_lrn, mw)  # not in archive
                home_rrn__archive_lrn__offsite_cn = _find_child_by_title(
                    home_rrn__archive_lrn, 'Offsite')
                
                # Children of LinkedResourceNode (a kind of _ResourceNode) that is inside ClusterNode
                # HACK: Relies on Crystal's "Not in Archive" page to provide a
                #       link to the original URL (which is offsite)
                await _expand_node(home_rrn__offsite_cn__twitter_lrn, mw)  # not in archive
                home_rrn__offsite_cn__twitter_lrn__offsite_cn = _find_child_by_title(
                    home_rrn__offsite_cn__twitter_lrn, 'Offsite')
                await _expand_node(home_rrn__embedded_cn__stylesheet_lrn)  # already downloaded
                home_rrn__embedded_cn__stylesheet_lrn__embedded_cn = _find_child_by_title(
                    home_rrn__embedded_cn__stylesheet_lrn, 'Embedded')
                
                # ---
                
                # Children of NormalResourceNode (a kind of _ResourceNode) that is inside ResourceGroupNode
                await _expand_node(feed_rgn__rss_feed_nrn)
                feed_rgn__rss_feed_nrn__home_rrn = _find_child(
                    feed_rgn__rss_feed_nrn, home_url, url_prefix)
            
            # Ensure suggested source for each kind of node in the Entity Tree is reasonable
            if True:
                # Children of RootNode
                assertEqual(None, await _source_name_for_node(home_rrn, mw))
                assertEqual(None, await _source_name_for_node(feed_rgn, mw))  # = source of Feed group
                assertEqual('Home', await _source_name_for_node(comic_rgn, mw))  # = source of Comic group
                
                # Children of RootResourceNode (a kind of _ResourceNode) that is inside RootNode
                assertEqual('Home', await _source_name_for_node(home_rrn__atom_feed_rrn, mw))
                assertEqual('Home', await _source_name_for_node(home_rrn__about_rrn, mw))
                assertEqual('Home', await _source_name_for_node(home_rrn__feed_glrn, mw))
                assertEqual('Home', await _source_name_for_node(home_rrn__feed_glrn__atom_feed_rrn, mw))
                assertEqual('Home', await _source_name_for_node(home_rrn__feed_glrn__rss_feed_lrn, mw))
                assertEqual('Home', await _source_name_for_node(home_rrn__archive_lrn, mw))
                assertEqual('Home', await _source_name_for_node(home_rrn__offsite_cn, mw))
                assertEqual('Home', await _source_name_for_node(home_rrn__offsite_cn__twitter_lrn, mw))
                assertEqual('Home', await _source_name_for_node(home_rrn__embedded_cn, mw))
                assertEqual('Home', await _source_name_for_node(home_rrn__embedded_cn__stylesheet_lrn, mw))
                
                # Children of ResourceGroupNode
                assertEqual(None, await _source_name_for_node(feed_rgn__atom_feed_rrn, mw))  # = source of Feed group
                assertEqual(None, await _source_name_for_node(feed_rgn__rss_feed_nrn, mw))  # = source of Feed group
                
                # ---
                
                # Children of LinkedResourceNode (a kind of _ResourceNode) that is inside GroupedLinkedResourceNode
                assertEqual(None, await _source_name_for_node(home_rrn__feed_glrn__rss_feed_lrn__home_rrn, mw))
                
                # Children of LinkedResourceNode (a kind of _ResourceNode) that is inside RootResourceNode
                assertEqual(None, await _source_name_for_node(home_rrn__archive_lrn__offsite_cn, mw))
                
                # Children of LinkedResourceNode (a kind of _ResourceNode) that is inside ClusterNode
                assertEqual(None, await _source_name_for_node(home_rrn__offsite_cn__twitter_lrn__offsite_cn, mw))
                assertEqual(None, await _source_name_for_node(home_rrn__embedded_cn__stylesheet_lrn__embedded_cn, mw))
                
                # ---
                
                # Children of NormalResourceNode (a kind of _ResourceNode) that is inside ResourceGroupNode
                assertEqual('Feed', await _source_name_for_node(feed_rgn__rss_feed_nrn__home_rrn, mw))


async def _expand_node(node_ti: TreeItem, mw: MainWindow | None=None, *, will_download: bool=False) -> None:
    node_ti.Expand()
    if will_download:
        if mw is None:
            raise ValueError('Need mw parameter when will_download=True')
        await wait_for_download_to_start_and_finish(
            mw.task_tree,
            stacklevel_extra=1)
    await wait_for(
        first_child_of_tree_item_is_not_loading_condition(node_ti),
        timeout=3.0,  # took 2.2s on Windows CI
        stacklevel_extra=1)


async def _source_name_for_node(node_ti: TreeItem, mw: MainWindow) -> str | None:
    node_ti.SelectItem()
    
    click_button(mw.new_group_button)
    ngd = await NewGroupDialog.wait_for()
    source_name = ngd.source  # capture
    await ngd.cancel()
    return source_name


# === Test: Preview Members ===

@skip('covered by: test_given_resource_node_with_multiple_link_children_matching_url_pattern_can_create_new_group_to_bundle_those_links_together')
async def test_given_pattern_contains_no_wildcards_then_preview_members_show_1_matching_url() -> None:
    pass


@skip('covered by: test_given_resource_node_with_multiple_link_children_matching_url_pattern_can_create_new_group_to_bundle_those_links_together')
async def test_given_pattern_contains_wildcards_then_preview_members_show_all_matching_urls() -> None:
    pass


async def test_given_urls_loaded_and_new_url_created_when_show_new_group_dialog_and_input_pattern_matching_new_url_then_preview_members_shows_new_url() -> None:
    with database_on_ssd(False), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        if True:
            home_url = sp.get_request_url('https://xkcd.com/')
            
            comic1_url = sp.get_request_url('https://xkcd.com/1/')
            comic2_url = sp.get_request_url('https://xkcd.com/2/')
            comic_pattern = sp.get_request_url('https://xkcd.com/#/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            # Create initial entities
            home_rr = RootResource(project, 'Home', Resource(project, home_url))
            ResourceGroup(project, 'Comic', comic_pattern, source=home_rr)
            
            root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
            home_ti = _find_child(root_ti, home_url)
            comic_ti = _find_child(root_ti, comic_pattern)
            
            # Expand Comic group, to formally load URLs
            # Prepare to spy on whether LoadUrlsProgressDialog appears
            with patch.object(
                    project._load_urls_progress_listener,
                    'will_load_resources',
                    wraps=project._load_urls_progress_listener.will_load_resources) as progress_listener_method:
                comic_ti.Expand()
                await wait_for(tree_item_has_no_children_condition(comic_ti))
                assert 0 == len(comic_ti.Children)
                
                # Ensure did show LoadUrlsProgressDialog
                assert progress_listener_method.call_count >= 1
            
            # Expand home URL, to discover initial comic URLs
            home_ti.Expand()
            await wait_for_download_to_start_and_finish(mw.task_tree)
            assert first_child_of_tree_item_is_not_loading_condition(home_ti)()
            assert len(comic_ti.Children) >= 2  # contains at least the first and last comics
            
            # Start creating a duplicate Comic group
            if True:
                assert mw.new_group_button.Enabled
                click_button(mw.new_group_button)
                ngd = await NewGroupDialog.wait_for()
                
                # Input new URL pattern with wildcard, to match comics
                ngd.pattern_field.Value = comic_pattern
                
                # Ensure preview members show the appropriate matching URLs
                member_urls = [
                    ngd.preview_members_list.GetString(i)
                    for i in range(ngd.preview_members_list.GetCount())
                ]
                assert comic1_url in member_urls  # contains first comic
                assert len(member_urls) >= 2  # contains last comic too
                
                await ngd.cancel()


# === Test: New Group Options ===

async def test_when_add_group_then_does_not_download_immediately_by_default() -> None:
    with _served_simple_site_with_group() as (home_url, comic_pattern):
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            rr = RootResource(project, 'Home', Resource(project, home_url))
            rr.download()
            await wait_for_download_to_start_and_finish(mw.task_tree)
            
            assert mw.new_group_button.Enabled
            click_button(mw.new_group_button)
            ngd = await NewGroupDialog.wait_for()
            
            ngd.pattern_field.Value = comic_pattern
            ngd.name_field.Value = 'Comic'
            await ngd.ok()
            
            # Ensure did NOT start downloading
            root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
            root_ti.find_child(comic_pattern, project.default_url_prefix)
            assert tree_has_no_children_condition(mw.task_tree)()


async def test_when_add_group_can_download_group_immediately_with_1_extra_click() -> None:
    with _served_simple_site_with_group() as (home_url, comic_pattern):
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            rr = RootResource(project, 'Home', Resource(project, home_url))
            rr.download()
            await wait_for_download_to_start_and_finish(mw.task_tree)
            
            assert mw.new_group_button.Enabled
            click_button(mw.new_group_button)
            ngd = await NewGroupDialog.wait_for()
            
            ngd.pattern_field.Value = comic_pattern
            ngd.name_field.Value = 'Comic'
            click_checkbox(ngd.download_immediately_checkbox)  # extra click #1
            await ngd.ok()
            
            # Ensure started downloading
            root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
            root_ti.find_child(comic_pattern, project.default_url_prefix)
            await wait_for_download_to_start_and_finish(mw.task_tree)


async def test_when_edit_group_then_new_group_options_not_shown() -> None:
    with _served_simple_site_with_group() as (_, comic_pattern):
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            ResourceGroup(project, 'Comic', comic_pattern)
            root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
            comic_ti = root_ti.find_child(comic_pattern, project.default_url_prefix)
            comic_ti.SelectItem()
            
            assert mw.edit_button.Enabled
            click_button(mw.edit_button)
            ngd = await NewGroupDialog.wait_for()
            
            assert not ngd.new_options_shown
            await ngd.cancel()


@contextmanager
def _served_simple_site_with_group() -> Iterator[tuple[str, str]]:
    server = MockHttpServer({
        '/': dict(
            status_code=200,
            headers=[('Content-Type', 'text/html')],
            content=b'<a href="/assets/image.png">Comic 1</a>'
        ),
        '/assets/image.png': dict(
            status_code=200,
            headers=[('Content-Type', 'image/png')],
            content=b''
        )
    })
    with server:
        home_url = server.get_url('/')
        comic_pattern = server.get_url('/**')
        
        yield (home_url, comic_pattern)


# === Utility ===

_find_child = TreeItem.find_child
_find_child_by_title = TreeItem.find_child_by_title

# NOTE: Only for use with tree items in EntityTree
_assert_tree_item_icon_tooltip_contains = EntityTree.assert_tree_item_icon_tooltip_contains



"""
Additional collections on top of the built-in Python collections and the
extended collections in the standard `collections` module.
"""

from collections import OrderedDict
import sys
from typing import cast, TypeVar


class defaultordereddict(OrderedDict):
    def __init__(self, default_factory=None):
        super().__init__()
        self.default_factory = default_factory
    
    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        value = self.default_factory()
        self[key] = value
        return value


_K = TypeVar('_K')
_V = TypeVar('_V')

def as_ordereddict(d: dict[_K, _V]) -> 'OrderedDict[_K, _V]':
    assert sys.version_info >= (3, 8)
    return cast('OrderedDict[_K, _V]', d)



import datetime


def datetime_is_aware(dt: datetime.datetime) -> bool:
    # https://docs.python.org/3/library/datetime.html#determining-if-an-object-is-aware-or-naive
    return (
        dt.tzinfo is not None and
        dt.tzinfo.utcoffset(dt) is not None
    )


def datetime_is_in_utc(dt: datetime.datetime) -> bool:
    return (
        dt.tzinfo is not None and
        dt.tzinfo.utcoffset(dt) == datetime.timedelta(0)
    )



from crystal.util.xos import (
    is_linux, is_mac_os, is_windows, windows_major_version,
)
import json
import os
import os.path
import subprocess


# NOTE: Neither the "ssd" nor "ssd_checker" projects on PyPI,
#       which are supposed to implement this function,
#       actually work on macOS, as of 2023-12-08.
def is_ssd(itempath: str) -> bool:
    """
    Returns whether the specified file is definitely on a local solid state disk.
    
    Such a disk is assumed to support O(1) random reads and have a high
    maximum IOPS.
    """
    if is_mac_os():
        return _is_mac_ssd(itempath)
    elif is_linux():
        return _is_linux_ssd(itempath)
    elif is_windows():
        return _is_windows_ssd(itempath)
    else:  # unknown OS
        return False


def _mountpoint(abs_itempath: str) -> str:
    assert abs_itempath.startswith('/')
    mountpoint_dirpath = abs_itempath
    while not os.path.ismount(mountpoint_dirpath):
        new_mountpoint_dirpath = os.path.dirname(mountpoint_dirpath)
        if mountpoint_dirpath == new_mountpoint_dirpath:
            raise AssertionError('Root directory is not considered a mount point: ' + mountpoint_dirpath)
        mountpoint_dirpath = new_mountpoint_dirpath  # reinterpret
    return mountpoint_dirpath


def _is_mac_ssd(itempath: str) -> bool:
    import plistlib
    
    mountpoint_dirpath = _mountpoint(os.path.realpath(itempath))
    mountpoint_info = plistlib.loads(subprocess.check_output(
        ['diskutil', 'info', '-plist', mountpoint_dirpath],
    ))
    return mountpoint_info.get('SolidState', False)


def _is_linux_ssd(itempath: str) -> bool:
    mountpoint_dirpath = _mountpoint(os.path.realpath(itempath))
    lsblk_info = json.loads(subprocess.check_output(
        ['lsblk', '-o', 'name,rota,mountpoints', '--list', '--json'],
    ).decode('utf-8'))
    for device in lsblk_info['blockdevices']:
        for device_mount in device['mountpoints']:
            if device_mount == mountpoint_dirpath:
                rota = device['rota']
                assert isinstance(rota, bool)
                return not rota
    return False  # mountpoint not found


def _is_windows_ssd(itempath: str) -> bool:
    import win32file

    itempath = os.path.realpath(itempath)
    # TODO: Does this handle UNC paths correctly?
    drive = os.path.splitdrive(itempath)[0].upper()
    drivetype = win32file.GetDriveType(drive)
    if drivetype == win32file.DRIVE_RAMDISK:
        return True
    elif drivetype in (win32file.DRIVE_FIXED, win32file.DRIVE_REMOVABLE):
        winver = windows_major_version()
        if winver is not None and winver < 8:
            # Need Windows 8+ to access MSFT_PhysicalDisk in WMI, according to:
            # https://learn.microsoft.com/en-us/windows-hardware/drivers/storage/msft-physicaldisk
            return False
        
        import wmi

        c = wmi.WMI()
        phy_to_part = 'Win32_DiskDriveToDiskPartition'
        log_to_part = 'Win32_LogicalDiskToPartition'
        index = {
            log_disk.Caption: phy_disk.Index
            for phy_disk in c.Win32_DiskDrive()
            for partition in phy_disk.associators(phy_to_part)
            for log_disk in partition.associators(log_to_part)
        }
        
        if drive in index:
            c = wmi.WMI(moniker='//./ROOT/Microsoft/Windows/Storage')
            return bool(c.MSFT_PhysicalDisk(DeviceId=str(index[drive]), MediaType=4))
        else:
            return False
    else:
        return False



"""
The startup module of Crystal, which is designated as the initial module
that runs when Crystal is launched.

Trampolines to the main module at crystal.main.

The startup module and the main module are kept separate because app
packages like py2app and py2exe treat the startup module slightly
differently than a regular module, so it's best not to put too much
functionality in the startup module. Known differences:
* py2app --
    Relocates the startup module from its original location at /crystal/__main__.py
    to just /__main__.py, so it's more difficult to patch in automated tests.
* py2exe --
    The startup module has no __file__ defined in its globals() and 
    doesn't appear to be patchable at all by automated tests.
"""

from crystal.main import main

if __name__ == '__main__':
    main()



"""
Provides services for downloading a ResourceRevision.
"""

from collections.abc import Iterable
from crystal import __version__
from crystal.model import (
    ProjectHasTooManyRevisionsError, Resource, ResourceRevision,
    ResourceRevisionMetadata,
)
from crystal.util.xos import is_mac_os, is_windows
from crystal.util.xthreading import fg_call_and_wait
from http.client import HTTPConnection, HTTPSConnection
import ssl
import truststore
from typing import BinaryIO, cast, Collection, Dict, TYPE_CHECKING
import urllib.error
from urllib.parse import urlparse
import urllib.request

if TYPE_CHECKING:
    from crystal.task import DownloadResourceBodyTask


HTTP_REQUEST_TIMEOUT = 10  # seconds

# The User-Agent string to use for downloads, or None to omit.
_USER_AGENT_STRING = 'Crystal/%s (https://dafoster.net/projects/crystal-web-archiver/)' % __version__

# Whether to log verbose output related to HTTP requests and responses.
# Can be used to inspect the exact request & response lines and headers exchanged.
# 
# For temporary experimentation in the shell.
# 
# TODO: Alter the format of the verbose output to be as easy to read as
#       curl's similar -v option.
_VERBOSE_HTTP_REQUESTS_AND_RESPONSES = False

# Extra HTTP headers to add to every request, overriding any standard headers.
# 
# For temporary experimentation in the shell.
# 
# WARNING: These headers will NOT be persisted in the downloaded ResourceRevision
#          and therefore it will not be possible to repeat a similar request in
#          the future to get a more up-to-date version of a prior revision.
_EXTRA_HEADERS = dict()  # type: Dict[str, str]


def download_resource_revision(
        resource: Resource,
        progress_listener: 'DownloadResourceBodyTask',
        ) -> ResourceRevision:
    """
    Synchronously downloads a revision of the specified resource.
    For internal use by DownloadResourceBodyTask.
    
    Arguments:
    * resource -- the resource to download.
    * progress_listener -- the DownloadResourceBodyTask that progress updates will be sent to.
    
    Raises:
    * ProjectHasTooManyRevisionsError
    """
    
    if resource.project.request_cookie_applies_to(resource.url):
        request_cookie = resource.project.request_cookie
    else:
        request_cookie = None
    
    if resource.definitely_has_no_revisions:
        known_etags = ()  # type: Collection[str]
    else:
        known_etags = fg_call_and_wait(lambda: resource.revision_for_etag().keys())
    
    try:
        progress_listener.subtitle = 'Waiting for response...'
        (metadata, body_stream) = ResourceRequest.create(
            resource.url, 
            request_cookie,
            known_etags,
        )()
        try:
            # If an HTTP 304 response was received without an ETag,
            # try to populate the ETag value so that future calls to
            # ResourceRevision.resolve_http_304() will actually be
            # able to find the original revision
            if metadata is not None and metadata['status_code'] == 304:
                response_etags = [
                    cur_value
                    for (cur_name, cur_value) in metadata['headers']
                    if cur_name.lower() == 'etag'
                ]
                response_etag = response_etags[0] if len(response_etags) > 0 else None
                if response_etag is None and len(known_etags) == 1:
                    (known_etag,) = known_etags
                    metadata['headers'].append(['ETag', known_etag])
            
            # TODO: Provide incremental feedback such as '7 KB of 15 KB'
            progress_listener.subtitle = 'Receiving response...'
            return ResourceRevision.create_from_response(
                resource,
                metadata,
                body_stream,
                request_cookie
            )
        finally:
            body_stream.close()
    except ProjectHasTooManyRevisionsError:
        raise
    except Exception as error:
        # TODO: Handle rare case where this raises an IO error when writing
        #       revision to the database, perhaps by creating/returning
        #       an *unsaved* ResourceRevision in memory.
        return ResourceRevision.create_from_error(
            resource,
            error,
            request_cookie
        )


class ResourceRequest:
    """
    Encapsulates a request to fetch a resource.
    """
    
    @staticmethod
    def create(
            url: str,
            request_cookie: str | None=None,
            known_etags: Iterable[str]=()
            ) -> 'ResourceRequest':
        """
        Raises:
        * urllib.error.URLError -- if URL scheme not supported.
        """
        url_parts = urlparse(url)
        if url_parts.scheme in ('http', 'https'):
            return HttpResourceRequest(url, request_cookie, known_etags)
        elif url_parts.scheme == 'ftp':
            return UrlResourceRequest(url)
        else:
            raise urllib.error.URLError('URL scheme "%s" is not supported.' % url_parts.scheme)
    
    def __call__(self) -> tuple[ResourceRevisionMetadata | None, BinaryIO]:
        """
        Returns a (metadata, body_stream) tuple, where
            `metadata` is a JSON-serializable dictionary or None and
            `body_stream` is a file-like object (which supports `read` and `close`).
        
        Raises any Exception.
        """
        raise NotImplementedError


class HttpResourceRequest(ResourceRequest):
    def __init__(self, 
            url: str,
            request_cookie: str | None=None,
            known_etags: Iterable[str]=()
            ) -> None:
        if urlparse(url).scheme not in ('http', 'https'):
            raise ValueError('Expected URL with http or https scheme')
        self.url = url
        self._request_cookie = request_cookie
        self._known_etags = known_etags
    
    def __call__(self) -> tuple[ResourceRevisionMetadata, BinaryIO]:
        url_parts = urlparse(self.url)
        scheme = url_parts.scheme
        host_and_port = url_parts.netloc
        path_and_query = (
            (url_parts.path or '/') + 
            ('' if url_parts.query == '' else f'?{url_parts.query}')
        )
        
        if scheme == 'http':
            conn = HTTPConnection(
                host_and_port,
                timeout=HTTP_REQUEST_TIMEOUT,
            )
        elif scheme == 'https':
            conn = HTTPSConnection(
                host_and_port,
                timeout=HTTP_REQUEST_TIMEOUT,
                context=get_ssl_context(),
            )
        else:
            raise ValueError('Not an HTTP(S) URL.')
        
        if _VERBOSE_HTTP_REQUESTS_AND_RESPONSES:
            conn.set_debuglevel(1)
        
        headers = {}
        if _USER_AGENT_STRING is not None:
            headers['User-Agent'] = _USER_AGENT_STRING
        if self._request_cookie is not None:
            headers['Cookie'] = self._request_cookie
        if_none_match_value = ', '.join(self._known_etags)
        if if_none_match_value != '':
            headers['If-None-Match'] = if_none_match_value
        for (k, v) in _EXTRA_HEADERS.items():
            headers[k] = v
        
        conn.request('GET', path_and_query, headers=headers)
        response = conn.getresponse()
        
        metadata = ResourceRevisionMetadata({
            'http_version': response.version,
            'status_code': response.status,
            'reason_phrase': response.reason,
            'headers': [[k, v] for (k, v) in response.getheaders()]
        })
        body_stream = _HttpResourceBodyStream(
            close=conn.close,
            read=response.read,
            readinto=response.readinto,
            fileno=response.fileno,
            mode='rb')
        return (metadata, cast(BinaryIO, body_stream))
    
    def __repr__(self):
        return 'HttpResourceRequest(%s)' % repr(self.url)


class _HttpResourceBodyStream:
    """
    File-like object for reading from an HTTP resource.
    """
    def __init__(self, close, read, readinto, fileno, mode) -> None:
        self.close = close
        self.read = read
        self.readinto = readinto
        self.fileno = fileno
        self.mode = mode


class UrlResourceRequest(ResourceRequest):
    def __init__(self, url):
        self.url = url
    
    def __call__(self):
        request = urllib.request.Request(self.url)
        response = urllib.request.urlopen(request, context=get_ssl_context())
        return (None, response)
    
    def __repr__(self):
        return 'UrlResourceRequest(%s)' % repr(self.url)


_SSL_CONTEXT = None

def get_ssl_context() -> ssl.SSLContext:
    """
    Creates the SSLContext used to make HTTPS connections.
    
    In particular, loads any CA certificates needed to authenticate those connections.
    """
    global _SSL_CONTEXT
    if _SSL_CONTEXT is None:
        # Load certifications from OS certificate store.
        # 
        # Optimize options to connect from client to server
        # (rather than optimizing for a server receiving a client connection),
        # emulating the behavior of ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH).
        # 
        # On Linux:
        # - OpenSSL's $SSL_CERT_FILE and $SSL_CERT_DIR environment variables
        #   can be used to override the default CA certificates.
        # - Certificate locations for major distros are supported,
        #   with the authoritative list in truststore/_openssl.py:
        #   https://github.com/sethmlarson/truststore/blob/8725ad5ea26966eeb2960be24b85ca6c9371f11f/src/truststore/_openssl.py#L8-L17
        ctx = truststore.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        ctx.verify_mode = ssl.CERT_REQUIRED
        ctx.check_hostname = True
        
        _SSL_CONTEXT = ctx  # export
    return _SSL_CONTEXT



"""
Tests that .crystalproj packages show/hide their file extension appropriately
in the desktop environment and in the main window title.
"""
from unittest import skip

# === macOS ===

@skip('not yet automated')
def test_given_macos_when_crystalproj_created_then_crystalproj_file_extension_is_hidden_in_finder(self) -> None:
    pass


@skip('not yet automated')
def test_given_macos_and_crystalproj_hides_file_extension_when_crystalproj_opened_then_crystalproj_file_extension_is_hidden_in_main_window_title(self) -> None:
    pass


@skip('not yet automated')
def test_given_macos_and_crystalproj_shows_file_extension_when_crystalproj_opened_as_read_only_then_crystalproj_file_extension_is_shown_in_main_window_title(self) -> None:
    pass


@skip('not yet automated')
def test_given_macos_when_crystalproj_opened_as_writable_then_crystalproj_file_extension_is_hidden_in_finder_and_is_hidden_in_main_window_title(self) -> None:
    pass


# === Windows ===

# TODO: Find way to hide .crystalproj extension in Windows
@skip('not yet automated')
def test_given_windows_when_crystalproj_created_then_crystalproj_file_extension_is_shown_in_explorer(self) -> None:
    pass


@skip('not yet automated')
def test_given_windows_and_crystalproj_shows_file_extension_when_crystalproj_opened_then_crystalproj_file_extension_is_shown_in_main_window_title(self) -> None:
    pass


# === Linux ===

# TODO: Find way to hide .crystalproj extension in Linux
@skip('not yet automated')
def test_given_linux_when_crystalproj_created_then_crystalproj_file_extension_is_shown_in_explorer(self) -> None:
    pass


@skip('not yet automated')
def test_given_linux_and_crystalproj_shows_file_extension_when_crystalproj_opened_then_crystalproj_file_extension_is_shown_in_main_window_title(self) -> None:
    pass



from ast import literal_eval
from collections.abc import Iterator
from contextlib import contextmanager
from crystal import __version__ as crystal_version
from crystal.tests.util.asserts import assertEqual, assertIn, assertNotIn
from crystal.tests.util.screenshots import take_error_screenshot
from crystal.tests.util.server import served_project
from crystal.tests.util.skip import skipTest
from crystal.tests.util.subtests import SubtestsContext, with_subtests
from crystal.tests.util.wait import (
    DEFAULT_WAIT_PERIOD, DEFAULT_WAIT_TIMEOUT, HARD_TIMEOUT_MULTIPLIER,
    wait_for_sync, WaitTimedOut,
)
from crystal.tests.util.windows import MainWindow
from crystal.util.xos import is_asan, is_windows
from crystal.util.xthreading import fg_call_and_wait
from io import TextIOBase
import os
import re
from select import select
import subprocess
import sys
import tempfile
import textwrap
import time
from typing import List, Optional
from unittest import skip
from unittest.mock import ANY
import urllib
import warnings

_EXPECTED_PROXY_PUBLIC_MEMBERS = []  # type: List[str]

_EXPECTED_PROJECT_PUBLIC_MEMBERS = [
    'FILE_EXTENSION',
    'OPENER_FILE_EXTENSION',
    'add_task',
    'close',
    'default_url_prefix',
    'get_display_url',
    'get_resource',
    'get_resource_group',
    'get_root_resource',
    'hibernate_tasks',
    'html_parser_type',
    'is_valid',
    'listeners',
    'load_urls',
    'major_version',
    'min_fetch_date',
    'path',
    'readonly',
    'request_cookie',
    'request_cookie_applies_to',
    'request_cookies_in_use',
    'resource_groups',
    'resources',
    'resources_matching_pattern',
    'root_resources',
    'root_task',
    'unhibernate_tasks',
    'urls_matching_pattern',
]

_EXPECTED_WINDOW_PUBLIC_MEMBERS = [
    'close',
    'entity_tree',
    'project',
    'start_server',
    'task_tree'
]

# ------------------------------------------------------------------------------
# Tests

@with_subtests
def test_can_launch_with_shell(subtests: SubtestsContext) -> None:
    with crystal_shell() as (crystal, banner):
        with subtests.test(msg='with informative banner'):
            # ...containing Crystal's version
            assertIn(f'Crystal {crystal_version}', banner)
            # ...containing the Python version
            python_version = '.'.join([str(x) for x in sys.version_info[:3]])
            assertIn(f'Python {python_version}', banner)
            # ...mentioning the "help" command
            assertIn('"help"', banner)
            # ...mentioning the "project" and "window" variables
            assertIn('"project"', banner)
            assertIn('"window"', banner)
            # ...mentioning how to exit (with both "exit" and Ctrl-D (or Ctrl-Z plus Return))
            assert (
                'Use exit() or Ctrl-D (i.e. EOF) to exit.' in banner or
                'Use exit() or Ctrl-Z plus Return to exit.' in banner
            ), banner
        
        with subtests.test(msg='and {project, window} can be used as placeholders, before main window appears'):
            assertEqual('<unset crystal.model.Project proxy>\n', _py_eval(crystal, 'project'))
            assertEqual('<unset crystal.browser.MainWindow proxy>\n', _py_eval(crystal, 'window'))
            
            assertIn('Help on _Proxy in module ', _py_eval(crystal, 'help(project)'))
            assertIn('Help on _Proxy in module ', _py_eval(crystal, 'help(window)'))
            
            # Ensure public members match expected set
            assertEqual(repr(_EXPECTED_PROXY_PUBLIC_MEMBERS) + '\n',
                _py_eval(crystal, "[x for x in dir(project) if not x.startswith('_')]"))
            assertEqual(repr(_EXPECTED_PROXY_PUBLIC_MEMBERS) + '\n',
                _py_eval(crystal, "[x for x in dir(window) if not x.startswith('_')]"))
        
        # Open MainWindow by creating new empty project
        _create_new_empty_project(crystal)
        
        with subtests.test(msg='and {project, window} can be used for real, after main window appears'):
            assert re.fullmatch(
                r'^<crystal\.model\.Project object at 0x[0-9a-f]+>\n$',
                _py_eval(crystal, 'project'))
            assert re.fullmatch(
                r'^<crystal\.browser\.MainWindow object at 0x[0-9a-f]+>\n$',
                _py_eval(crystal, 'window'))
        
        with subtests.test(msg='and {project, window} can be used with help()'):
            assertIn('Help on Project in module crystal.model object:', _py_eval(crystal, 'help(project)'))
            assertIn('Help on MainWindow in module crystal.browser object:', _py_eval(crystal, 'help(window)'))


def test_can_use_pythonstartup_file() -> None:
    with tempfile.NamedTemporaryFile(suffix='.py') as startup_file:
        startup_file.write(textwrap.dedent('''\
            EXCLUDED_URLS = ['a', 'b', 'c']
            '''
        ).encode('utf-8'))
        startup_file.flush()
        
        with crystal_shell(env_extra={'PYTHONSTARTUP': startup_file.name}) as (crystal, _):
            assertEqual(
                ['a', 'b', 'c'],
                literal_eval(_py_eval(crystal, 'EXCLUDED_URLS')),
                'Expected variables written at top-level by '
                    '$PYTHONSTARTUP file to be accessible from shell'
            )


# NOTE: This test code was split out of the test_can_launch_with_shell() test above
#       because it is particularly easy to break and having a separate test function
#       makes the break type quicker to identify.
@with_subtests
def test_builtin_globals_have_stable_public_api(subtests: SubtestsContext) -> None:
    with crystal_shell() as (crystal, _):
        # Open MainWindow by creating new empty project
        _create_new_empty_project(crystal)
        
        with subtests.test(global_name='project'):
            assertEqual(_EXPECTED_PROJECT_PUBLIC_MEMBERS,
                literal_eval(_py_eval(crystal, "[x for x in dir(project) if not x.startswith('_')]")),
                'Public API of Project class has changed')
        
        with subtests.test(global_name='window'):
            assertEqual(_EXPECTED_WINDOW_PUBLIC_MEMBERS,
                literal_eval(_py_eval(crystal, "[x for x in dir(window) if not x.startswith('_')]")),
                'Public API of MainWindow class has changed')


@with_subtests
def test_shell_exits_with_expected_message(subtests: SubtestsContext) -> None:
    _ensure_can_use_crystal_shell()
    
    with subtests.test(case='test when first open/create dialog is closed given shell is running then shell remains running'):
        with crystal_shell() as (crystal, _):
            _close_open_or_create_dialog(crystal)
            
            try:
                assertEqual('4\n', _py_eval(crystal, '2 + 2'))
            except AssertionError as e:
                raise AssertionError(f'{e} Trailing output: {_drain(crystal)!r}')
    
    with subtests.test(case='test when main window or non-first open/create dialog is closed given shell is running then shell remains running'):
        with crystal_shell() as (crystal, _):
            _create_new_empty_project(crystal)
            _close_main_window(crystal)
            
            _close_open_or_create_dialog(crystal)
            
            try:
                assertEqual(
                    '4\n',
                    _py_eval(crystal, '2 + 2', timeout=5.0))  # took >4.0s in Linux CI
            except AssertionError as e:
                raise AssertionError(f'{e} Trailing output: {_drain(crystal)!r}')
    
    for exit_method in ('exit()', 'Ctrl-D'):
        with subtests.test(case=f'test when {exit_method} given first open/create dialog is already closed then exits'):
            with crystal_shell() as (crystal, _):
                assert isinstance(crystal.stdin, TextIOBase)
                
                _close_open_or_create_dialog(crystal)
                
                if exit_method == 'exit()':
                    _py_eval(crystal, 'exit()', stop_suffix='')
                elif exit_method == 'Ctrl-D':
                    crystal.stdin.close()  # Ctrl-D
                else:
                    raise AssertionError()
                
                _wait_for_crystal_to_exit(
                    crystal,
                    timeout=DEFAULT_WAIT_TIMEOUT)
        
        with subtests.test(case=f'test when {exit_method} given non-first open/create dialog is already closed then exits'):
            with crystal_shell() as (crystal, _):
                assert isinstance(crystal.stdin, TextIOBase)
                
                _create_new_empty_project(crystal)
                _close_main_window(crystal)
                
                _close_open_or_create_dialog(crystal)
                
                if exit_method == 'exit()':
                    _py_eval(
                        crystal, 'exit()', stop_suffix='',
                        timeout=5.0)  # took 4.0s in Linux CI
                elif exit_method == 'Ctrl-D':
                    crystal.stdin.close()  # Ctrl-D
                else:
                    raise AssertionError()
                
                _wait_for_crystal_to_exit(
                    crystal,
                    timeout=5.0)  # took >4.0s in Linux CI
    
    for exit_method in ('exit()', 'Ctrl-D'):
        with subtests.test(case=f'test when {exit_method} given first open/create dialog still open then prints waiting message and does not exit'):
            with crystal_shell() as (crystal, _):
                assert isinstance(crystal.stdin, TextIOBase)
                
                _close_open_or_create_dialog(crystal, after_delay=.5)
                
                if exit_method == 'exit()':
                    _py_eval(crystal, 'exit()', stop_suffix='now waiting for all windows to close...\n')
                elif exit_method == 'Ctrl-D':
                    crystal.stdin.close()  # Ctrl-D
                    _read_until(crystal.stdout, 'now waiting for all windows to close...\n')
                else:
                    raise AssertionError()
                
                _wait_for_crystal_to_exit(
                    crystal,
                    timeout=.5 + DEFAULT_WAIT_TIMEOUT)
        
        with subtests.test(case=f'test when {exit_method} given main window still open then prints waiting message and does not exit'):
            with crystal_shell() as (crystal, _):
                assert isinstance(crystal.stdin, TextIOBase)
                
                _create_new_empty_project(crystal)
                
                _close_main_window(crystal, after_delay=.5)
                _close_open_or_create_dialog(crystal, after_delay=.5*2)
                
                if exit_method == 'exit()':
                    _py_eval(crystal, 'exit()', stop_suffix='now waiting for all windows to close...\n')
                elif exit_method == 'Ctrl-D':
                    crystal.stdin.close()  # Ctrl-D
                    _read_until(crystal.stdout, 'now waiting for all windows to close...\n')
                else:
                    raise AssertionError()
                
                _wait_for_crystal_to_exit(
                    crystal,
                    timeout=.5*2 + DEFAULT_WAIT_TIMEOUT)
        
        with subtests.test(case=f'test when {exit_method} given non-first open/create dialog still open then prints waiting message and does not exit'):
            with crystal_shell() as (crystal, _):
                assert isinstance(crystal.stdin, TextIOBase)
                
                _create_new_empty_project(crystal)
                _close_main_window(crystal)
                
                _close_open_or_create_dialog(crystal, after_delay=.5)
                
                if exit_method == 'exit()':
                    _py_eval(crystal, 'exit()', stop_suffix='now waiting for all windows to close...\n')
                elif exit_method == 'Ctrl-D':
                    crystal.stdin.close()  # Ctrl-D
                    _read_until(crystal.stdout, 'now waiting for all windows to close...\n')
                else:
                    raise AssertionError()
                
                _wait_for_crystal_to_exit(
                    crystal,
                    timeout=.5 + DEFAULT_WAIT_TIMEOUT)


def test_when_typed_code_raises_exception_then_print_traceback() -> None:
    with crystal_shell() as (crystal, _):
        expected_traceback = (
            'Traceback (most recent call last):\n'
            '  File "<console>", line 1, in <module>\n'
            'NameError: name \'Resource\' is not defined\n'
        )
        assertEqual(expected_traceback, _py_eval(crystal, 'Resource'))


@with_subtests
def test_can_read_project_with_shell(subtests: SubtestsContext) -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = 'https://xkcd.com/'
        
        project_path = sp.project.path  # capture
        fg_call_and_wait(lambda: sp.project.close())
        
        with crystal_shell() as (crystal, _):
            with subtests.test(case='test can open project', return_if_failure=True):
                # Test can import Project
                assertEqual('', _py_eval(crystal, 'from crystal.model import Project'))
                # Test can open project
                assertEqual('', _py_eval(crystal, f'p = Project({sp.project.path!r})'))
            
            with subtests.test(case='test can list project entities'):
                assertEqual(
                    "[RootResource('Home','https://xkcd.com/')]\n",
                    _py_eval(crystal, 'list(p.root_resources)[:1]'))
                assertEqual(
                    "[ResourceGroup('Comics','https://xkcd.com/#/')]\n",
                    _py_eval(crystal, 'list(p.resource_groups)'))
                assertEqual(
                    '71\n',
                    _py_eval(crystal, 'len(p.resources)'))
                assertEqual(
                    "Resource('https://xkcd.com/')\n",
                    _py_eval(crystal, 'list(p.resources)[0]'))
            
            with subtests.test(case='test can get project entities', return_if_failure=True):
                assertEqual(
                    "Resource('https://xkcd.com/')\n",
                    _py_eval(crystal, f'r = p.get_resource({home_url!r}); r'))
                
                assertEqual(
                    "[<ResourceRevision 1 for 'https://xkcd.com/'>]\n",
                    _py_eval(crystal, f'list(r.revisions())'))
                assertEqual(
                    "<ResourceRevision 1 for 'https://xkcd.com/'>\n",
                    _py_eval(crystal, f'rr = r.default_revision(); rr'))
                
                assertEqual(
                    "RootResource('Home','https://xkcd.com/')\n",
                    _py_eval(crystal, f'root_r = p.get_root_resource(r); root_r'))
                
                assertEqual(
                    "ResourceGroup('Comics','https://xkcd.com/#/')\n",
                    _py_eval(crystal, f'rg = p.get_resource_group("Comics"); rg'))
                assertEqual(
                    '14\n',
                    _py_eval(crystal, f'len(rg.members)'))
                assertEqual(
                    "Resource('https://xkcd.com/1/')\n",
                    _py_eval(crystal, f'list(rg.members)[0]'))
            
            with subtests.test(case='test can read content of resource revision'):
                assertEqual(
                    {
                        'http_version': 11,
                        'status_code': 200,
                        'reason_phrase': 'OK',
                        'headers': ANY
                    },
                    literal_eval(_py_eval(crystal, f'rr.metadata')))
                _py_eval(crystal, f'with rr.open() as f:\n    body = f.read()\n', stop_suffix='>>> ')
                assertEqual(
                    r"""b'<!DOCTYPE html>\n<html>\n<head>\n<link rel="stylesheet" type="text/css" href="/s/7d94e0.css" title="Default"/>\n<title>xkcd: Air Gap</title>\n'""" + '\n',
                    _py_eval(crystal, f'body[:137]'))
            
            with subtests.test(case='test can serve resource revision'):
                # Test can import ProjectServer
                assertEqual('', _py_eval(crystal, 'from crystal.server import ProjectServer'))
                assertEqual('', _py_eval(crystal, 'from io import StringIO'))
                # Test can start ProjectServer
                assertEqual(
                    "",
                    _py_eval(
                        crystal, f'server = ProjectServer(p, stdout=StringIO())',
                        timeout=8.0  # 2.0s and 4.0s isn't long enough for macOS test runners on GitHub Actions
                    ))
                port = literal_eval(_py_eval(crystal, f'server.port'))
                request_url = literal_eval(_py_eval(crystal, f'server.get_request_url({home_url!r})'))
                
                # Test ProjectServer serves resource revision
                assertIn(str(port), request_url)
                with urllib.request.urlopen(request_url) as response:
                    response_bytes = response.read()
                assertIn(b'<title>xkcd: Air Gap</title>', response_bytes)


@with_subtests
def test_can_write_project_with_shell(subtests: SubtestsContext) -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        if True:
            home_url = sp.get_request_url('https://xkcd.com/')
            
            comic1_url = sp.get_request_url('https://xkcd.com/1/')
            comic2_url = sp.get_request_url('https://xkcd.com/2/')
            comic_pattern = sp.get_request_url('https://xkcd.com/#/')
            
            atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
            rss_feed_url = sp.get_request_url('https://xkcd.com/rss.xml')
            feed_pattern = sp.get_request_url('https://xkcd.com/*.xml')
        
        # Create named temporary directory that won't be deleted automatically
        with tempfile.NamedTemporaryFile(suffix='.crystalproj', delete=False) as project_td:
            pass
        os.remove(project_td.name)
        project_dirpath = project_td.name
        
        with crystal_shell() as (crystal, _):
            with subtests.test(case='test can create project', return_if_failure=True):
                # Test can import Project
                assertEqual('', _py_eval(crystal, 'from crystal.model import Project'))
                # Test can create project
                assertEqual('', _py_eval(crystal, f'p = Project({project_dirpath!r})',
                    # 2.0s isn't long enough for macOS test runners on GitHub Actions
                    timeout=4.0))
            
            with subtests.test(case='test can create project entities', return_if_failure=True):
                # Test can import Resource
                assertEqual('', _py_eval(crystal, 'from crystal.model import Resource'))
                # Test can create Resource
                assertEqual(
                    "Resource('http://localhost:2798/_/https/xkcd.com/')\n",
                    _py_eval(crystal, f'r = Resource(p, {home_url!r}); r'))
                
                # Test can import RootResource
                assertEqual('', _py_eval(crystal, 'from crystal.model import RootResource'))
                # Test can create RootResource
                assertEqual(
                    "RootResource('Home','http://localhost:2798/_/https/xkcd.com/')\n",
                    _py_eval(crystal, f'root_r = RootResource(p, "Home", r); root_r'))
                
                # Test can download ResourceRevision
                with _delay_between_downloads_minimized(crystal):
                    assertEqual('', _py_eval(crystal, 'rr_future = r.download()'))
                    # TODO: Use wait_for_sync() rather than a manual loop
                    while True:
                        is_done = (literal_eval(_py_eval(crystal, 'rr_future.done()')) == True)
                        if is_done:
                            break
                        time.sleep(.2)
                    assertIn('<ResourceRevision ', _py_eval(crystal, 'rr = rr_future.result(); rr'))
                
                # Test can import ResourceGroup
                assertEqual('', _py_eval(crystal, 'from crystal.model import ResourceGroup'))
                # Test can create ResourceGroup
                assertEqual(
                    "ResourceGroup('Comic','http://localhost:2798/_/https/xkcd.com/#/')\n",
                    _py_eval(crystal, f'rg = ResourceGroup(p, "Comic", {comic_pattern!r}); rg'))
                # Ensure ResourceGroup includes some members discovered by downloading resource Home
                def rg_member_count() -> int:
                    count = literal_eval(_py_eval(crystal, f'len(rg.members)'))
                    assert isinstance(count, int)
                    return count
                wait_for_sync(lambda: 9 == rg_member_count())
            
            with subtests.test(case='test can delete project entities', return_if_failure=True):
                # Test can delete ResourceGroup
                assertEqual('', _py_eval(crystal, f'rg_m = list(rg.members)[0]'))
                assertEqual('', _py_eval(crystal, f'rg.delete()'))
                # Ensure ResourceGroup itself is deleted
                assertEqual('', _py_eval(crystal, f'p.get_resource_group(rg.name)'))
                # Ensure former members of ResourceGroup still exist
                assertEqual('True\n', _py_eval(crystal, f'p.get_resource(rg_m.url) == rg_m'))
                
                # Test can delete RootResource
                assertEqual('', _py_eval(crystal, f'root_r_r = root_r.resource'))
                assertEqual('', _py_eval(crystal, f'root_r.delete()'))
                # Ensure RootResource itself is deleted
                assertEqual('', _py_eval(crystal, f'p.get_root_resource(root_r_r)'))
                # Ensure former target of RootResource still exists
                assertEqual('True\n', _py_eval(crystal, f'p.get_resource(root_r_r.url) == root_r_r'))
                
                # Test can delete ResourceRevision
                assertEqual('', _py_eval(crystal, f'rr_r = rr.resource'))
                assertEqual('1\n', _py_eval(crystal, f'len(list(rr_r.revisions()))'))
                assertEqual('', _py_eval(crystal, f'rr.delete()'))
                # Ensure ResourceRevision itself is deleted
                assertEqual('0\n', _py_eval(crystal, f'len(list(rr_r.revisions()))'))
                
                # Test can delete Resource
                assertEqual('', _py_eval(crystal, f'r.delete()'))
                # Ensure Resource itself is deleted
                assertEqual('', _py_eval(crystal, f'p.get_resource(r.url)'))
            
            with subtests.test(case='test can download project entities', return_if_failure=True):
                # Recreate home Resource
                assertEqual(
                    "Resource('http://localhost:2798/_/https/xkcd.com/')\n",
                    _py_eval(crystal, f'r = Resource(p, {home_url!r}); r'))
                # Recreate home RootResource
                assertEqual(
                    "RootResource('Home','http://localhost:2798/_/https/xkcd.com/')\n",
                    _py_eval(crystal, f'root_r = RootResource(p, "Home", r); root_r'))
                
                # Test can download RootResource
                with _delay_between_downloads_minimized(crystal):
                    assertEqual('', _py_eval(crystal, 'rr_future = root_r.download()'))
                    # TODO: Use wait_for_sync() rather than a manual loop
                    while True:
                        is_done = (literal_eval(_py_eval(crystal, 'rr_future.done()')) == True)
                        if is_done:
                            break
                        time.sleep(.2)
                    assertIn('<ResourceRevision ', _py_eval(crystal, 'rr = rr_future.result(); rr'))
                
                # Create feed ResourceGroup
                assertEqual(
                    "ResourceGroup('Feed','http://localhost:2798/_/https/xkcd.com/*.xml')\n",
                    _py_eval(crystal, f'rg = ResourceGroup(p, "Feed", {feed_pattern!r}); rg'))
                assertEqual(
                    "",
                    _py_eval(crystal, f'rg.source = root_r'))
                # Ensure ResourceGroup includes some members discovered by downloading resource Home
                assertEqual(
                    2,
                    literal_eval(_py_eval(crystal, f'len(rg.members)')))
                
                # Test can download ResourceGroup
                with _delay_between_downloads_minimized(crystal):
                    assertEqual('', _py_eval(crystal, 'drgt = rg.download()'))
                    # TODO: Use wait_for_sync() rather than a manual loop
                    while True:
                        is_done = (literal_eval(_py_eval(crystal, 'drgt.complete')) == True)
                        if is_done:
                            break
                        time.sleep(.2)
                assertEqual(
                    [True] * 2,
                    literal_eval(_py_eval(crystal, '[r.has_any_revisions() for r in rg.members]')))


@skip('covered by: test_can_write_project_with_shell')
def test_can_open_or_create_project() -> None:
    pass


@skip('covered by: test_can_write_project_with_shell')
def test_can_create_project_entities() -> None:
    pass


@skip('covered by: test_can_read_project_with_shell')
def test_can_read_project_entities() -> None:
    pass


@skip('covered by: test_can_write_project_with_shell')
def test_can_download_project_entities() -> None:
    pass


@skip('covered by: test_can_write_project_with_shell')
def test_can_delete_project_entities() -> None:
    pass


def test_can_import_guppy_in_shell() -> None:
    with crystal_shell() as (crystal, _):
        # Ensure can import guppy
        import_result = literal_eval(_py_eval(crystal, 'import guppy; guppy.__version__'))
        assert isinstance(import_result, str)
        
        # Ensure can create hpy instance
        assertEqual('', _py_eval(crystal, 'from guppy import hpy; h = hpy()'))
        
        # Ensure can take memory sample
        result = _py_eval(crystal, 'import gc; gc.collect(); heap = h.heap(); heap; _.more')
        assertNotIn('Traceback', result)
        
        # Ensure can create checkpoint
        assertEqual('', _py_eval(crystal, 'h.setref()'))
        
        # Ensure can take memory sample since checkpoint
        result = _py_eval(crystal, 'import gc; gc.collect(); heap = h.heap(); heap; _.more')
        assertNotIn('Traceback', result)


# ------------------------------------------------------------------------------
# Utility: Windows

_OK_THREAD_STOP_SUFFIX = (
    # If thread finishes after the "t.start()" fully completes and writes the next '>>> ' prompt
    'OK\n',
    # If thread finishes before the "t.start()" fully completes and writes the next '>>> ' prompt
    'OK\n>>> ',
    # If the next '>>> ' prompt is written in the middle of "OK" + "\n" being printed.
    'OK>>> \n',
    # TODO: Determine how this empiricially observed situation is possible
    'OK\n>>> >>> ',
)

def _create_new_empty_project(crystal: subprocess.Popen) -> None:
    # NOTE: Uses private API, including the entire crystal.tests package
    _py_eval(crystal, textwrap.dedent('''\
        if True:
            from crystal.tests.util.runner import run_test
            from crystal.tests.util.windows import OpenOrCreateDialog
            import os
            import tempfile
            from threading import Thread
            #
            async def create_new_project():
                # Create named temporary directory that won't be deleted automatically
                with tempfile.NamedTemporaryFile(suffix='.crystalproj', delete=False) as project_td:
                    pass
                os.remove(project_td.name)
                os.mkdir(project_td.name)
                project_dirpath = project_td.name
                #
                ocd = await OpenOrCreateDialog.wait_for()
                mw = await ocd.create_and_leave_open(project_dirpath)
                #
                return mw
            #
            result_cell = [Ellipsis]
            def get_result(result_cell):
                result_cell[0] = run_test(lambda: create_new_project())
                print('OK')
            #
            t = Thread(target=lambda: get_result(result_cell))
            t.start()
        '''),
        stop_suffix=_OK_THREAD_STOP_SUFFIX,
        # NOTE: 6.0 was observed to sometimes not be long enough on macOS
        timeout=8.0
    )
    assertEqual(
        "<class 'crystal.tests.util.windows.MainWindow'>\n",
        _py_eval(crystal, 'type(result_cell[0])'))


def _close_open_or_create_dialog(crystal: subprocess.Popen, *, after_delay: float | None=None) -> None:
    # NOTE: Uses private API, including the entire crystal.tests package
    _py_eval(crystal, textwrap.dedent(f'''\
        if True:
            from crystal.tests.util.runner import bg_sleep, run_test
            from crystal.tests.util.windows import OpenOrCreateDialog
            from threading import Thread
            #
            async def close_ocd():
                ocd = await OpenOrCreateDialog.wait_for()
                if {after_delay} != None:
                    await bg_sleep({after_delay})
                ocd.open_or_create_project_dialog.Close()
                #
                print('OK')
            #
            t = Thread(target=lambda: run_test(close_ocd))
            t.start()
        '''), stop_suffix=_OK_THREAD_STOP_SUFFIX if after_delay is None else '')


def _close_main_window(crystal: subprocess.Popen, *, after_delay: float | None=None) -> None:
    # NOTE: Uses private API, including the entire crystal.tests package
    _py_eval(crystal, textwrap.dedent(f'''\
        if True:
            from crystal.tests.util.runner import bg_sleep, run_test
            from crystal.tests.util.windows import MainWindow
            from threading import Thread
            #
            async def close_main_window():
                mw = await MainWindow.wait_for()
                if {after_delay} != None:
                    await bg_sleep({after_delay})
                await mw.close()
                #
                print('OK')
            #
            t = Thread(target=lambda: run_test(close_main_window))
            t.start()
        '''),
        stop_suffix=_OK_THREAD_STOP_SUFFIX if after_delay is None else '',
        timeout=MainWindow.CLOSE_TIMEOUT,
    )


@contextmanager
def _delay_between_downloads_minimized(crystal: subprocess.Popen) -> Iterator[None]:
    # NOTE: Uses private API, including the entire crystal.tests package
    _py_eval(crystal, 'from crystal.tests.util.downloads import delay_between_downloads_minimized as D')
    _py_eval(crystal, 'download_ctx = D()')
    _py_eval(crystal, 'download_ctx.__enter__()')
    try:
        yield
    finally:
        _py_eval(crystal, 'download_ctx.__exit__(None, None, None)')


# ------------------------------------------------------------------------------
# Utility: Shell

@contextmanager
def crystal_shell(*, env_extra={}) -> Iterator[tuple[subprocess.Popen, str]]:
    """
    Context which starts "crystal --shell" upon enter
    and cleans up the associated process upon exit.
    """
    _ensure_can_use_crystal_shell()
    
    # Determine how to run Crystal on command line
    crystal_command: list[str]
    python = sys.executable
    if getattr(sys, 'frozen', None) == 'macosx_app':
        python_neighbors = os.listdir(os.path.dirname(python))
        (crystal_binary_name,) = (n for n in python_neighbors if 'crystal' in n.lower())
        crystal_binary = os.path.join(os.path.dirname(python), crystal_binary_name)
        crystal_command = [crystal_binary]
    else:
        crystal_command = [python, '-m', 'crystal']
    
    crystal = subprocess.Popen(
        [*crystal_command, '--shell'],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        encoding='utf-8',
        env={
            **os.environ,
            **{
                'CRYSTAL_FAULTHANDLER': 'True',
                
                # Prevent Python warnings from being mixed into output
                'PYTHONWARNINGS': 'ignore',
                
                # Prevent profiling warnings from being mixed into output
                # TODO: Recommend renaming to avoid (double-)negatives
                'CRYSTAL_NO_PROFILE_FG_TASKS': 'True',
                'CRYSTAL_NO_PROFILE_GC': 'True',
                'CRYSTAL_NO_PROFILE_RECORD_LINKS': 'True',
            },
            **env_extra
        })
    try:
        assert isinstance(crystal.stdout, TextIOBase)
        
        (banner, _) = _read_until(
            crystal.stdout, '\n>>> ',
            timeout=4.0  # 2.0s isn't long enough for macOS test runners on GitHub Actions
        )
        yield (crystal, banner)
    finally:
        assert crystal.stdin is not None
        crystal.stdin.close()
        assert crystal.stdout is not None
        crystal.stdout.close()
        crystal.kill()
        crystal.wait()


def _ensure_can_use_crystal_shell() -> None:
    if is_windows():
        # NOTE: Windows doesn't provide stdout for graphical processes,
        #       which is needed by the current implementation.
        #       Workaround is possible with run_exe.py but time-consuming to implement.
        skipTest('not supported on Windows; graphical subprocesses are mute')
    if is_asan():
        # NOTE: ASan slows down many operations, causing shell operations to
        #       spuriously fail timeout checks, even when
        #       CRYSTAL_GLOBAL_TIMEOUT_MULTIPLIER is used
        skipTest('too slow when run with Address Sanitizer')


def _py_eval(
        python: subprocess.Popen,
        py_code: str,
        stop_suffix: str | tuple[str, ...] | None=None,
        *, timeout: float | None=None) -> str:
    if '\n' in py_code and stop_suffix is None:
        raise ValueError(
            'Unsafe to use _py_eval() on multi-line py_code '
            'unless stop_suffix is set carefully')
    if stop_suffix is None:
        stop_suffix = '>>> '
    
    assert isinstance(python.stdin, TextIOBase)
    assert isinstance(python.stdout, TextIOBase)
    python.stdin.write(f'{py_code}\n'); python.stdin.flush()
    (result, found_stop_suffix) = _read_until(
        python.stdout, stop_suffix, timeout=timeout, stacklevel_extra=1)
    return result[:-len(found_stop_suffix)]


def _read_until(
        stream: TextIOBase,
        stop_suffix: str | tuple[str, ...],
        timeout: float | None=None,
        *, period: float | None=None,
        stacklevel_extra: int=0
        ) -> tuple[str, str]:
    """
    Reads from the specified stream until the provided `stop_suffix`
    is read at the end of the stream or the timeout expires.
    
    Raises:
    * ReadUntilTimedOut -- if the timeout expires before `stop_suffix` is read
    """
    if isinstance(stop_suffix, str):
        stop_suffix = (stop_suffix,)
    if timeout is None:
        timeout = DEFAULT_WAIT_TIMEOUT
    if period is None:
        period = DEFAULT_WAIT_PERIOD
    
    soft_timeout = timeout
    hard_timeout = timeout * HARD_TIMEOUT_MULTIPLIER
    
    stop_suffix_bytes_choices = [s.encode(stream.encoding) for s in stop_suffix]
    
    read_buffer = b''
    found_stop_suffix = None  # type: Optional[str]
    start_time = time.time()
    hard_timeout_exceeded = False
    try:
        delta_time = 0.0
        while True:
            # 1. Wait for stream to be ready to read or hit EOF
            # 2. Read stream (if became ready)
            # 3. Look for an acceptable `stop_suffix` at the end of everything read so far
            remaining_time = hard_timeout - delta_time
            assert remaining_time > 0
            (rlist_ready, _, xlist_ready) = select(
                [stream],
                [],
                [stream],
                remaining_time)
            did_time_out = (len(rlist_ready) + len(xlist_ready) == 0)
            if did_time_out:
                # Look for an acceptable "stop suffix"
                # Special case: '' is always an acceptable `stop_suffix`
                if '' in stop_suffix:
                    found_stop_suffix = ''
            else:
                # Read stream
                # NOTE: Append uses quadratic performance.
                #       Not using for large amounts of text so I don't care.
                read_buffer += stream.buffer.read1(1024)  # type: ignore[attr-defined]  # arbitrary
                
                # Look for an acceptable "stop suffix"
                for (i, s_bytes) in enumerate(stop_suffix_bytes_choices):
                    if read_buffer.endswith(s_bytes):
                        found_stop_suffix = stop_suffix[i]
                        break
            if found_stop_suffix is not None:
                # Done
                break
            
            # If hard timeout exceeded then raise
            delta_time = time.time() - start_time
            if did_time_out or delta_time >= hard_timeout:
                # Screenshot the timeout error
                take_error_screenshot()
                
                hard_timeout_exceeded = True
                read_so_far = read_buffer.decode(stream.encoding)
                raise ReadUntilTimedOut(
                    f'Timed out after {timeout:.1f}s while '
                    f'reading until {stop_suffix!r}. '
                    f'Read so far: {read_so_far!r}',
                    read_so_far)
    finally:
        # If soft timeout exceeded then warn before returning
        if not hard_timeout_exceeded:
            delta_time = time.time() - start_time
            if delta_time > soft_timeout:
                warnings.warn(
                    ('Soft timeout exceeded (%.1fs > %.1fs) while '
                    'reading until %r.') % (
                        delta_time,
                        soft_timeout,
                        stop_suffix
                    ),
                    stacklevel=(2 + stacklevel_extra))
            
    return (read_buffer.decode(stream.encoding), found_stop_suffix)


class ReadUntilTimedOut(WaitTimedOut):
    def __init__(self, message: str, read_so_far: str) -> None:
        super().__init__(message)
        self.read_so_far = read_so_far


_DEFAULT_DRAIN_TTL = min(DEFAULT_WAIT_TIMEOUT, 2.0)

def _drain(stream: TextIOBase | subprocess.Popen, ttl: float | None=None) -> str:
    """
    Reads as much as possible from the specified stream for the specified
    TTL duration and returns it.
    
    Often useful for debugging _read_until() failures.
    """
    if ttl is None:
        ttl = _DEFAULT_DRAIN_TTL
    
    if isinstance(stream, subprocess.Popen):
        stream2 = stream.stdout
        assert isinstance(stream2, TextIOBase)
        stream = stream2  # reinterpret
    
    EOT = '\4'  # End of Transmission; an unlikely character to occur in the wild
    try:
        _read_until(stream, EOT, ttl)
    except ReadUntilTimedOut as e:
        return e.read_so_far
    else:
        raise ValueError('Actually encountered EOT while reading stream!')


# ------------------------------------------------------------------------------
# Utility: Wait

def _wait_for_crystal_to_exit(
        crystal: subprocess.Popen,
        *, timeout: float,
        stacklevel_extra: int=0
        ) -> None:
    start_time = time.time()  # capture
    try:
        crystal.wait(timeout=timeout)
    except subprocess.TimeoutExpired:
        try:
            crystal.wait(timeout=(timeout * HARD_TIMEOUT_MULTIPLIER) - timeout)
        except subprocess.TimeoutExpired:
            raise WaitTimedOut('Timed out waiting for Crystal to exit') from None
        else:
            delta_time = time.time() - start_time
            warnings.warn(
                ('Soft timeout exceeded (%.1fs > %.1fs) while '
                'waiting for Crystal to exit') % (
                    delta_time,
                    timeout,
                ),
                stacklevel=(2 + stacklevel_extra))


# ------------------------------------------------------------------------------



from __future__ import annotations

from collections.abc import Callable, Iterator
from contextlib import contextmanager
from crystal.tests.util.runner import pump_wx_events
from crystal.util.wx_treeitem_gettooltip import GetTooltipEvent
from crystal.util.xos import is_windows
from typing import List, Literal, Optional
from unittest.mock import patch
import wx

# ------------------------------------------------------------------------------
# Utility: Controls: wx.Button

def click_button(button: wx.Button) -> None:
    # Dispatch wx.EVT_BUTTON event
    event = wx.PyCommandEvent(wx.EVT_BUTTON.typeId, button.GetId())
    event.SetEventObject(button)
    assert event.GetEventObject().GetId() == button.GetId()
    button.Command(event)


# ------------------------------------------------------------------------------
# Utility: Controls: wx.CheckBox

def set_checkbox_value(checkbox: wx.CheckBox, value: bool) -> None:
    """
    Changes the value of a checkbox in a way that fires a realistic
    wx.EVT_CHECKBOX event if appropriate.
    """
    if checkbox.Value != value:
        click_checkbox(checkbox)
        assert checkbox.Value == value


def click_checkbox(checkbox: wx.CheckBox) -> None:
    old_value = checkbox.Value  # capture
    
    # Toggle value
    checkbox.Value = not checkbox.Value
    
    # Dispatch wx.EVT_CHECKBOX event
    event = wx.PyCommandEvent(wx.EVT_CHECKBOX.typeId, checkbox.GetId())
    event.SetEventObject(checkbox)
    assert event.GetEventObject().GetId() == checkbox.GetId()
    checkbox.ProcessEvent(event)
    
    new_value = checkbox.Value  # capture
    assert new_value != old_value, 'Expected checkbox to toggle value'


# ------------------------------------------------------------------------------
# Utility: Controls: wx.FileDialog

@contextmanager
def file_dialog_returning(filepath: str) -> Iterator[None]:
    with patch('wx.FileDialog', spec=True) as MockFileDialog:
        instance = MockFileDialog.return_value
        instance.ShowModal.return_value = wx.ID_OK
        instance.GetPath.return_value = filepath
        
        yield


# ------------------------------------------------------------------------------
# Utility: Controls: wx.Menu

def select_menuitem_now(menu: wx.Menu, menuitem_id: int) -> None:
    # Process the related wx.EVT_MENU event immediately,
    # so that the event handler is called before the wx.Menu is disposed
    event = wx.MenuEvent(type=wx.EVT_MENU.typeId, id=menuitem_id, menu=None)
    menu.ProcessEvent(event)


# ------------------------------------------------------------------------------
# Utility: Controls: wx.TreeCtrl

def get_children_of_tree_item(tree: wx.TreeCtrl, tii: wx.TreeItemId) -> list[TreeItem]:
    children = []  # type: List[TreeItem]
    next_child_tii = tree.GetFirstChild(tii)[0]
    while next_child_tii.IsOk():
        children.append(TreeItem(tree, next_child_tii))
        next_child_tii = tree.GetNextSibling(next_child_tii)  # reinterpret
    return children


class TreeItem:
    __slots__ = ['tree', 'id']
    
    _USE_FAST_ID_COMPARISONS = True
    
    def __init__(self, tree: wx.TreeCtrl, id: wx.TreeItemId) -> None:
        if not id.IsOk():
            raise ValueError('TreeItemId is invalid')
        
        self.tree = tree
        self.id = id
    
    # === Peer Queries and Actions ===
    
    @property
    def Text(self) -> str:
        return self.tree.GetItemText(self.id)
    
    @property
    def TextColour(self) -> wx.Colour:
        return self.tree.GetItemTextColour(self.id)
    
    @property
    def Bold(self) -> bool:
        return self.tree.IsBold(self.id)
    
    def Tooltip(self, tooltip_type: Literal['icon', 'label', None]=None) -> str | None:
        event = GetTooltipEvent(tree_item_id=self.id, tooltip_cell=[Ellipsis], tooltip_type=tooltip_type)
        self.tree.ProcessEvent(event)  # callee should set: event.tooltip_cell[0]
        assert event.tooltip_cell[0] is not Ellipsis
        return event.tooltip_cell[0]
    
    def SelectItem(self) -> None:
        self.tree.SelectItem(self.id)
    
    def IsSelected(self) -> bool:
        return self.tree.IsSelected(self.id)
    
    @staticmethod
    def GetSelection(tree: wx.TreeCtrl) -> TreeItem | None:
        selected_tii = tree.GetSelection()
        if selected_tii.IsOk():
            return TreeItem(tree, selected_tii)
        else:
            return None
    
    def Expand(self) -> None:
        self.tree.Expand(self.id)
    
    def Collapse(self) -> None:
        self.tree.Collapse(self.id)
    
    def IsExpanded(self) -> bool:
        return self.tree.IsExpanded(self.id)
    
    def ScrollTo(self) -> None:
        self.tree.ScrollTo(self.id)
    
    @staticmethod
    def GetRootItem(tree: wx.TreeCtrl) -> TreeItem:
        root_tii = tree.GetRootItem()
        assert root_tii.IsOk()
        return TreeItem(tree, root_tii)
    
    def GetFirstChild(self) -> TreeItem | None:
        first_child_tii = self.tree.GetFirstChild(self.id)[0]
        if first_child_tii.IsOk():
            return TreeItem(self.tree, first_child_tii)
        else:
            return None
    
    @property
    def Children(self) -> list[TreeItem]:
        return get_children_of_tree_item(self.tree, self.id)
    
    @property
    def _ItemData(self) -> object:
        return self.tree.GetItemData(self.id)
    
    # === Entity Tree: Find Child ===
    
    def find_child(
            parent_ti: TreeItem,
            url_or_url_pattern: str,
            default_url_prefix: str | None=None
            ) -> TreeItem:
        """
        Returns the first child of the specified parent tree item with the
        specified URL or URL pattern.
        
        Raises TreeItem.ChildNotFound if such a child is not found.
        """
        if default_url_prefix is not None:
            if url_or_url_pattern.startswith(default_url_prefix):
                url_or_url_pattern = url_or_url_pattern[len(default_url_prefix):]  # reinterpret
        try:
            (matching_child_ti,) = (
                child for child in parent_ti.Children
                if child.Text.startswith(f'{url_or_url_pattern} - ')
            )
        except ValueError:
            try:
                (matching_child_ti,) = (
                    child for child in parent_ti.Children
                    if child.Text == url_or_url_pattern
                )
            except ValueError:
                raise TreeItem.ChildNotFound(
                    f'Child {url_or_url_pattern} not found in specified TreeItem'
                ) from None
        return matching_child_ti
    
    def try_find_child(parent_ti: TreeItem, *args, **kwargs) -> TreeItem | None:
        try:
            return parent_ti.find_child(*args, **kwargs)
        except TreeItem.ChildNotFound:
            return None

    def find_child_by_title(parent_ti: TreeItem, title_fragment: str) -> TreeItem:
        """
        Returns the first child of the specified parent tree item whose
        title contains the specified fragment.
        
        Raises TreeItem.ChildNotFound if such a child is not found.
        """
        try:
            (matching_child_ti,) = (
                child for child in parent_ti.Children
                if title_fragment in child.Text
            )
            return matching_child_ti
        except ValueError:
            raise TreeItem.ChildNotFound(
                f'Child with title fragment {title_fragment!r} not found in specified TreeItem'
            ) from None
    
    class ChildNotFound(AssertionError):
        pass
    
    # === Operations ===
    
    async def right_click_showing_popup_menu(self, show_popup_menu: Callable[[wx.Menu], None]) -> None:
        raised_exc = None  # type: Optional[Exception]
        def PopupMenu(menu: wx.Menu, *args, **kwargs) -> bool:
            nonlocal raised_exc
            PopupMenu.called = True  # type: ignore[attr-defined]
            try:
                show_popup_menu(menu)
            except Exception as e:
                raised_exc = e
            return True
        PopupMenu.called = False  # type: ignore[attr-defined]
        with patch.object(self.tree, 'PopupMenu', PopupMenu):
            await self.right_click()
        assert PopupMenu.called  # type: ignore[attr-defined]
        if raised_exc is not None:
            raise raised_exc
    
    async def right_click(self) -> None:
        # TODO: Consider calling ProcessEvent directly rather than relying
        #       on pump_wx_events(), which has been observed to be unreliable
        #       on Windows in the past for posting individual events
        wx.PostEvent(self.tree, wx.TreeEvent(wx.EVT_TREE_ITEM_RIGHT_CLICK.typeId, self.tree, self.id))
        await pump_wx_events()
    
    # === Comparison ===
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TreeItem):
            return False
        if not (self.tree == other.tree):
            return False
        if TreeItem._USE_FAST_ID_COMPARISONS and not is_windows():
            # NOTE: wx.TreeItemId does not support equality comparison on Windows
            return self.id == other.id
        else:
            id_data = self._ItemData
            other_id_data = other._ItemData
            if id_data is None or other_id_data is None:
                raise TreeItemsIncomparableError('Cannot compare TreeItems lacking item data')
            return id_data is other_id_data


class TreeItemsIncomparableError(ValueError):
    pass


# ------------------------------------------------------------------------------



from __future__ import annotations

import asyncio
from collections.abc import Awaitable, Callable, Generator
from crystal.util.xthreading import bg_affinity, fg_affinity, fg_call_and_wait
import sys
import time
import traceback
from types import coroutine, FrameType
from typing import Generic, TYPE_CHECKING, TypeVar, Union
import urllib.error
import urllib.request

if TYPE_CHECKING:
    from crystal.tests.util.server import WebPage


_T = TypeVar('_T')


# ------------------------------------------------------------------------------
# Test Runner

@bg_affinity
def run_test(test_func: Callable[[], Awaitable[_T]] | Callable[[], _T]) -> _T:
    """
    Runs the specified test function.
    
    If the test function is async then it is run on the foreground thread.
    
    If the test function is sync then it is run on the (current) background thread.
    """
    test_co = test_func()  # if async func then should be a Generator[Command, None, _T]
    if not asyncio.iscoroutine(test_co):
        return test_co  # type: ignore[return-value]
    last_command_result = None  # type: Union[object, Exception]
    while True:
        try:
            command = fg_call_and_wait(
                lambda: test_co.send(last_command_result),  # type: ignore[attr-defined, union-attr]
                profile=False
            )
        except StopIteration as e:
            return e.value
        if not isinstance(command, Command):
            raise ValueError(
                'Async test function did yield something that was '
                f'not a Command: {command!r}')
        try:
            last_command_result = command.run()
        except Exception as e:
            last_command_result = e


@coroutine
@fg_affinity
def bg_sleep(  # type: ignore[misc]  # ignore non-Generator return type here
        duration: float
        ) -> Generator[Command, object, None]:
    """
    Switch to a background thread, sleep for the specified duration (in seconds), and
    then resume this foreground thread.
    """
    none_or_error = yield SleepCommand(duration)
    if none_or_error is None:
        return
    elif isinstance(none_or_error, Exception):
        raise none_or_error
    else:
        raise AssertionError()


@coroutine
@fg_affinity
def bg_fetch_url(
        url: str,
        *, headers: dict[str, str] | None=None,
        timeout: float | None=None,
        ) -> Generator[Command, object, WebPage]:
    """
    Switch to a background thread, fetch the specified URL, and
    then resume this foreground thread.
    """
    from crystal.tests.util.server import WebPage
    from crystal.tests.util.wait import DEFAULT_WAIT_TIMEOUT
    
    if timeout is None:
        timeout = DEFAULT_WAIT_TIMEOUT
    
    page_or_error = yield FetchUrlCommand(url, headers, timeout)
    if isinstance(page_or_error, WebPage):
        return page_or_error
    elif isinstance(page_or_error, Exception):
        raise page_or_error
    else:
        raise AssertionError()


@coroutine
@fg_affinity
def pump_wx_events() -> Generator[Command, object, None]:
    """
    Process all pending events on the wx event queue.
    """
    yield PumpWxEventsCommand()


@coroutine
@fg_affinity
def bg_breakpoint() -> Generator[Command, object, None]:
    """
    Stops the program in the debugger, with the foreground thread released,
    allowing the user to interact with the UI while the debugger has the program paused.
    
    Example usage:
        from crystal.tests.util.runner import bg_breakpoint
        await bg_breakpoint()
    """
    yield BreakpointCommand(sys._getframe(1))


class Command(Generic[_T]):  # abstract
    def run(self) -> _T:
        raise NotImplementedError()


class SleepCommand(Command[None]):
    def __init__(self, delay: float) -> None:
        self.delay = delay  # in seconds
    
    @bg_affinity
    def run(self) -> None:
        time.sleep(self.delay)


class FetchUrlCommand(Command['WebPage']):
    def __init__(self, url: str, headers: dict[str, str] | None, timeout: float) -> None:
        self._url = url
        self._headers = headers
        self._timeout = timeout
    
    @bg_affinity
    def run(self) -> WebPage:
        from crystal.tests.util.server import WebPage
        
        try:
            response_stream = urllib.request.urlopen(
                urllib.request.Request(
                    self._url,
                    headers=(self._headers or {})
                ),
                timeout=self._timeout)
        except urllib.error.HTTPError as e:
            response_stream = e
        with response_stream as response:
            response_bytes = response.read()
        return WebPage(response_stream.status, response_stream.headers, response_bytes)


class PumpWxEventsCommand(Command[None]):
    @bg_affinity
    def run(self) -> None:
        fg_call_and_wait(lambda: None)


class BreakpointCommand(Command[None]):
    def __init__(self, frame: FrameType) -> None:
        self._frame = frame
    
    @bg_affinity
    def run(self) -> None:
        print('Breakpoint hit at:')
        traceback.print_stack(self._frame)
        
        import pdb
        pdb.set_trace()


# ------------------------------------------------------------------------------



from crystal.util.xos import is_windows, windows_major_version


def decorate_label(icon: str, message: str, truncation_fix: str) -> str:
    if truncation_fix not in ('', ' ', '  '):
        raise ValueError()
    if is_windows():
        if windows_major_version() == 7:
            # Windows 7: Cannot display Unicode icon at all
            return f'{message}'
        else:  # windows_major_version() >= 8
            # Windows 8, 10: Can display some Unicode icons.
            # Some icons require trailing spaces to avoid truncating the message
            return f'{icon} {message}{truncation_fix}'
    return f'{icon} {message}'



"""
Parses HTML documents.
"""

from __future__ import annotations

from crystal.doc.generic import Document, Link
from typing import BinaryIO, Literal, Tuple

# HTML parsing library to use. See comparison between options at: 
# https://beautiful-soup-4.readthedocs.io/en/latest/#installing-a-parser
HtmlParserType = Literal['lxml', 'html_parser']

HTML_PARSER_TYPE_CHOICES = (
    HtmlParserType.__args__  # type: ignore[attr-defined]
)  # type: Tuple[HtmlParserType, ...]


def parse_html_and_links(
        html_bytes: bytes | BinaryIO,
        declared_charset: str | None,
        parser_type: HtmlParserType,
        ) -> tuple[Document, list[Link]] | None:
    """
    Parses the specified HTML bytestring, returning a 2-tuple containing
    (1) the HTML document and
    (2) a list of mutable links.
    
    The HTML document can be reoutput by getting its str() representation.
    
    Each link has the following mutable properties:
    * relative_url : str -- URL or URI referenced by this link, often relative.
    * type_title : str -- displayed title for this link's type.
    * title : str -- displayed title for this link, or None.
    * embedded : bool -- whether this link refers to an embedded resource.
    
    Arguments:
    * html_bytes -- HTML bytestring or file object.
    * declared_charset -- the encoding that the HTML document is declared to be in.
    """    
    import crystal.doc.html.basic as basic
    import crystal.doc.html.soup as soup

    # Convert html_bytes to string
    if hasattr(html_bytes, 'read'):
        html_bytes = html_bytes.read()  # type: ignore[union-attr]
    assert isinstance(html_bytes, bytes)
    
    # HACK: The BeautifulSoup parser doesn't currently handle <frameset>
    #       tags correctly. So workaround with a basic parser.
    if (b'frameset' in html_bytes) or (b'FRAMESET' in html_bytes):
        return basic.parse_html_and_links(html_bytes, declared_charset)
    else:
        return soup.parse_html_and_links(
            html_bytes, declared_charset, parser_type)



"""
Parses CSS documents.
"""

from collections.abc import Iterable
from crystal.doc.generic import Document, Link
import tinycss2
from tinycss2 import ast
from tinycss2.serializer import serialize_string_value, serialize_url
from typing import List, Tuple


def parse_css_and_links(
        body_bytes: bytes, 
        declared_charset: str | None=None
        ) -> 'Tuple[CssDocument, List[Link]]':
    (rules, _) = tinycss2.parse_stylesheet_bytes(
        body_bytes,
        protocol_encoding=declared_charset)
    return _parse_css_and_links(rules)


def parse_css_and_links_from_style_tag(
        tag_str: str
        ) -> 'Tuple[CssDocument, List[Link]]':
    rules = tinycss2.parse_stylesheet(tag_str)
    return _parse_css_and_links(rules)


def parse_css_and_links_from_style_attribute(
        attr_value: str
        ) -> 'Tuple[CssDocument, List[Link]]':
    decls = tinycss2.parse_declaration_list(attr_value)
    
    links = []  # type: List[Link]
    for decl in decls:
        if not isinstance(decl, ast.Declaration):
            continue
        _parse_links_from_component_values(decl.value, links)
    
    return (CssDocument(decls), links)


def _parse_css_and_links(rules) -> 'Tuple[CssDocument, List[Link]]':
    links = []  # type: List[Link]
    for rule in rules:
        if isinstance(rule, ast.QualifiedRule) or isinstance(rule, ast.AtRule):
            if rule.content is not None:  # has been observed as None in the wild sometimes
                _parse_links_from_component_values(rule.content, links)
        
        # @import "**";
        if isinstance(rule, ast.AtRule) and rule.at_keyword == 'import':
            for token in rule.prelude:
                if isinstance(token, ast.StringToken):
                    links.append(StringTokenLink(token))
    
    return (CssDocument(rules), links)


# https://doc.courtbouillon.org/tinycss2/stable/api_reference.html#term-component-values
def _parse_links_from_component_values(tokens, links: list[Link]) -> None:
    for token in tokens:
        # url(**)
        if isinstance(token, ast.URLToken):
            links.append(UrlTokenLink(token))
        
        # url("**")
        elif isinstance(token, ast.FunctionBlock):
            if (token.lower_name == 'url' and 
                    len(token.arguments) == 1 and 
                    isinstance(token.arguments[0], ast.StringToken)):
                links.append(UrlFunctionLink(token.arguments[0]))


class CssDocument(Document):
    def __init__(self, nodes: Iterable[ast.Node]) -> None:
        self._nodes = nodes
    
    def __str__(self) -> str:
        return tinycss2.serialize([
            n for n in self._nodes
            if not isinstance(n, ast.ParseError)
        ])


class UrlTokenLink(Link):
    def __init__(self, token: ast.URLToken) -> None:
        self._token = token
        
        self.title = None
        self.type_title = 'CSS URL Reference'
        self.embedded = True
    
    def _get_relative_url(self) -> str:
        return self._token.value
    def _set_relative_url(self, url: str) -> None:
        self._token.value = url
        self._token.representation = 'url(%s)' % serialize_url(url)
    relative_url = property(_get_relative_url, _set_relative_url)


class UrlFunctionLink(Link):
    def __init__(self, token: ast.StringToken) -> None:
        self._token = token
        
        self.title = None
        self.type_title = 'CSS URL Reference'
        self.embedded = True
    
    def _get_relative_url(self) -> str:
        return self._token.value
    def _set_relative_url(self, url: str) -> None:
        self._token.value = url
        self._token.representation = serialize_string_value(url)
    relative_url = property(_get_relative_url, _set_relative_url)


class StringTokenLink(Link):
    def __init__(self, token: ast.StringToken) -> None:
        self._token = token
        
        self.title = None
        self.type_title = 'CSS @import'
        self.embedded = True
    
    def _get_relative_url(self) -> str:
        return self._token.value
    def _set_relative_url(self, url: str) -> None:
        self._token.value = url
        self._token.representation = '"%s"' % serialize_string_value(url)
    relative_url = property(_get_relative_url, _set_relative_url)



import sqlite3


def is_database_closed_error(e: object) -> bool:
    return (
        isinstance(e, sqlite3.ProgrammingError) and 
        str(e) == 'Cannot operate on a closed database.'
    )


# Empirically, this error was observed to occur on macOS 15.3 when the disk
# containing the database was unmounted forcefully while the database was open.
def is_database_gone_error(e: object) -> bool:
    return (
        isinstance(e, sqlite3.DatabaseError) and
        str(e) == 'database disk image is malformed'
    )


# HACK: Some Windows runners in GitHub Actions don't have a version of SQLite
#       compiled with JSON for some reason. So some behavior must be altered
#       in that case.
def sqlite_has_json_support() -> bool:
    with sqlite3.connect(':memory:') as db:
        c = db.cursor()
        try:
            return list(c.execute('select json(1)')) == [('1',)]
        except sqlite3.OperationalError as e:
            if str(e) == 'no such function: json':
                return False
            else:
                raise



from collections.abc import Callable
from crystal.browser.icons import add_transparent_left_border
from crystal.util.wx_bind import bind
from crystal.util.xos import is_mac_os, is_windows
from typing import List
import wx


class Action:
    """
    Represents a command that can be accessed by multiple controls.
    
    When a property of an Action (ex: whether it is enabled) is changed,
    all associated controls will be updated automatically.
    """
    
    def __init__(self,
            menuitem_id: int=wx.ID_ANY,
            # You can prefix a letter in the label with & to underline it
            # and make it triggerable with Alt-<Letter> on Windows.
            # Linux and macOS will ignore & prefixes.
            label: str='',
            accel: wx.AcceleratorEntry | None=None,
            action_func: Callable[[wx.CommandEvent], None] | None=None,
            enabled: bool=True,
            button_bitmap: wx.Bitmap | None=None,
            button_label: str=''):
        self._menuitem_id = menuitem_id
        self._label = label
        self._accel = accel
        self._action_func = action_func
        self._enabled = enabled
        self._button_bitmap = button_bitmap
        self._button_label = button_label
        
        self._menuitems = []  # type: List[wx.MenuItem]
        self._buttons = []  # type: List[wx.Button]
    
    # === Properties ===
    
    def _get_enabled(self) -> bool:
        return self._enabled
    def _set_enabled(self, enabled: bool) -> None:
        self._enabled = enabled
        for mi in self._menuitems:
            mi.Enabled = enabled
        for b in self._buttons:
            b.Enabled = enabled
        
    enabled = property(_get_enabled, _set_enabled)
    
    # === Operations ===
    
    def append_menuitem_to(self, menu: wx.Menu) -> wx.MenuItem:
        """
        Creates a new wx.MenuItem associated with this action
        at the end of the specified menu.
        """
        menuitem = menu.Append(self._menuitem_id, self._label)
        if self._accel is not None:
            menuitem.Accel = self._accel
        if self._action_func is not None:
            bind(menu, wx.EVT_MENU, self._on_menuitem_command)
        menuitem.Enabled = self._enabled
        
        self._menuitems.append(menuitem)  # save reference
        return menuitem
    
    def create_button(self, *args, **kwargs) -> wx.Button:
        """
        Creates a new wx.Button associated with this action.
        """
        button_label = self._button_label or self._label
        if is_mac_os():
            # Don't use &<Letter> on macOS because it:
            # (1) creates an automatic ⌘<Letter> accelerator that, 
            #     when triggered, leaves the associated button stuck in a
            #     weird "pressed" state, and because it
            # (2) doesn't underline the letter or give any visual indication
            #     that the shortcut exists.
            # On macOS rely on menuitem accelerators instead.
            button_label = wx.Control.RemoveMnemonics(button_label)  # reinterpret
        
        button = wx.Button(*args, label=button_label, **kwargs)
        if self._button_bitmap is not None:
            button.SetBitmap(add_transparent_left_border(
                self._button_bitmap,
                # Alter left margin for bitmap to be reasonable on Windows.
                # Initially it is 0.
                8+6 if is_windows() else 0
            ))
            if is_windows():
                # Decrease bitmap-to-text spacing on Windows to be reasonable
                button.SetBitmapMargins((-6, 0))
        # (NOTE: Do NOT set accelerator for the button because it does not
        #        work consistently on macOS. Any action added as a button
        #        should also be added as a menuitem too, and we CAN set an
        #        accelerator reliably there.)
        if self._action_func is not None:
            bind(button, wx.EVT_BUTTON, self._action_func)
        button.Enabled = self._enabled
        
        self._buttons.append(button)  # save reference
        return button
    
    def dispose(self) -> None:
        # Avoid interacting with any further wx.Objects when this Action
        # is being deleted, because some of them are likely to have been 
        # deleted already, and such interactions could cause a crash
        self._menuitems.clear()
        self._buttons.clear()
    
    # === Events ===
    
    def _on_menuitem_command(self, event: wx.CommandEvent) -> None:
        if self._action_func is not None and event.Id in [m.Id for m in self._menuitems]:
            # NOTE: It is the responsiblity of the action function to call
            #       event.Skip() if appropriate.
            self._action_func(event)
        else:
            event.Skip()



from collections.abc import Awaitable, Callable, Iterator
from contextlib import contextmanager
from functools import wraps
from io import StringIO
import traceback
from unittest import SkipTest

# ------------------------------------------------------------------------------
# Utility: Subtests

class SubtestsContext:
    def __init__(self, test_name: str) -> None:
        self._test_name = test_name
        self._report = StringIO()
    
    # TODO: Allow with-statement to be used on SubtestsContext directly,
    #       eliminating the need to call run() in the API.
    @contextmanager
    def run(self) -> 'Iterator[SubtestsContext]':
        raised_exc = True
        try:
            yield self
            raised_exc = False
        except _SubtestReturn:
            raised_exc = False
        finally:
            subtest_report = self._report.getvalue()
            self._report.close()
            if len(subtest_report) != 0:
                print(subtest_report, end='')
                print('-' * 70)
                if not raised_exc:
                    raise SubtestFailed()
    
    @contextmanager
    def test(self, msg: str | None=None, **kwargs: object) -> Iterator[None]:
        """
        Context in which a subtest runs.
        
        The subtest is in failure output by its `msg`, `kwargs`, or both.
        
        If an exception is raised within a subtest context,
        by default the context is exited but the parent test continues
        executing after the context. If you want the parent test to
        exit instead, pass return_if_failure=True.
        
        Arguments:
        * msg --
            (Optional) Identifies this subtest in output.
        * return_if_failure --
            (Optional) Whether to return if an exception is raised in this context.
            Defaults to False.
        * kwargs --
            (Optional) Identifies this subtest in output.
        """
        if msg is None and len(kwargs) == 0:
            raise ValueError()
        return_if_failure = bool(kwargs.pop('return_if_failure', False))
        
        try:
            yield
        except Exception as e:
            if isinstance(e, AssertionError):
                exc_category = 'FAILURE'
                exc_traceback_useful = True
            elif isinstance(e, SkipTest):
                exc_category = 'SKIP'
                exc_traceback_useful = False
            else:
                exc_category = 'ERROR'
                exc_traceback_useful = True
            
            subtest_name_parts = [f'[{msg}] '] if msg is not None else []
            for (k, v) in kwargs.items():
                subtest_name_parts.append(f'({k}={v!r}) ')
            subtest_name = ''.join(subtest_name_parts).rstrip()
            
            print('- ' * (70 // 2), file=self._report)
            print(f'SUBTEST: {self._test_name} {subtest_name}', file=self._report)
            print('. ' * (70 // 2), file=self._report)
            if exc_traceback_useful:
                traceback.print_exc(file=self._report)
            print(exc_category, file=self._report)
            
            if return_if_failure:
                raise _SubtestReturn()
        else:
            # Passed. No output.
            pass


class _SubtestReturn(BaseException):
    pass


class SubtestFailed(Exception):
    def __init__(self) -> None:
        super().__init__('Subtest failed')


def with_subtests(test_func: Callable[[SubtestsContext], None]) -> Callable[[], None]:
    """Decorates a test function which can use subtests."""
    test_func_id = (test_func.__module__, test_func.__name__)
    test_name = f'{test_func_id[0]}.{test_func_id[1]}'
    
    subtests = SubtestsContext(test_name)
    
    @wraps(test_func)
    def wrapper():
        with subtests.run():
            test_func(subtests)
    return wrapper


def awith_subtests(test_func: Callable[[SubtestsContext], Awaitable[None]]) -> Callable[[], Awaitable[None]]:
    """Decorates a test function which can use subtests."""
    test_func_id = (test_func.__module__, test_func.__name__)
    test_name = f'{test_func_id[0]}.{test_func_id[1]}'
    
    subtests = SubtestsContext(test_name)
    
    @wraps(test_func)
    async def wrapper():
        raised_exc = True
        try:
            await test_func(subtests)
            raised_exc = False
        except _SubtestReturn:
            raised_exc = False
        finally:
            subtest_report = subtests._report.getvalue()
            if len(subtest_report) != 0:
                print(subtest_report, end='')
                print('-' * 70)
                if not raised_exc:
                    raise Exception('Subtests did fail')
    return wrapper


# ------------------------------------------------------------------------------


from crystal.util.wx_dialog import position_dialog_initially, ShowModal
import time
from typing import List, Optional, overload, Self, TypeVar
from typing_extensions import override
import wx

_DELAY_UNTIL_PROGRESS_DIALOG_SHOWS = 100 / 1000  # sec


# ------------------------------------------------------------------------------
# _AbstractProgressDialog

class _AbstractProgressDialog:
    _dialog_title: str  # abstract
    _CancelException: type[Exception]  # abstract
    
    _dialog_style: int | None
    _dialog: 'Optional[DeferredProgressDialog]'
    
    def __init__(self) -> None:
        self._dialog_style = None
        self._dialog = None
        self.reset()
    
    def reset(self) -> None:
        """
        Resets this progress listener to its initial state,
        hiding any progress dialogs.
        
        Afterward, the progress of a new operation may be
        reported through this same progress listener.
        """
        self._dialog_style = None
        if self._dialog is not None:
            self._dialog.Destroy()
            self._dialog = None  # very important; avoids segfaults
    
    # === Utility ===
    
    # protected
    def _update_can_cancel(self, can_cancel: bool, new_message: str) -> None:
        """
        Updates whether the visible progress dialog shows a cancel button or not,
        recreating the dialog with/without the button as necessary.
        
        Also updates the progress dialog to display the specified message.
        
        Does NOT raise _CancelException.
        """
        old_style = self._dialog_style
        new_style = (
            wx.PD_AUTO_HIDE|wx.PD_APP_MODAL
            if not can_cancel
            else wx.PD_AUTO_HIDE|wx.PD_APP_MODAL|wx.PD_CAN_ABORT|wx.PD_ELAPSED_TIME
        )
        new_name = 'cr-opening-project'
        
        if new_style != old_style:
            if self._dialog is not None:
                self._dialog.Destroy()
                self._dialog = None  # very important; avoids segfaults
            
            self._dialog_style = new_style
            self._dialog = DeferredProgressDialog(
                self._dialog_title,
                # NOTE: Message must be non-empty to size dialog correctly on Windows
                new_message,
                # (TODO: Shouldn't the value of the previous dialog version,
                #        if any, be preserved here?)
                # TODO: Shouldn't the maximum of the previous dialog version,
                #       if any, be preserved here?
                maximum=1,
                style=new_style,
                # Show dialog soon, only if it didn't complete in the meantime
                show_after=_DELAY_UNTIL_PROGRESS_DIALOG_SHOWS,
            )
            self._dialog.Name = new_name
        else:
            assert self._dialog is not None
            try:
                self._update(self._dialog.Value, new_message)
            except self._CancelException:
                # Ignore cancel request
                pass
    
    # protected
    def _update(self, new_value: int, new_message: str='') -> None:
        """
        Updates the value of the progress bar in the visible progress dialog.
        
        Optionally also updates the message in the progress dialog as well.
        
        Raises:
        * _CancelException
        """
        assert self._dialog is not None
        (ok, _) = self._dialog.Update(new_value, new_message)
        if not ok:
            raise self._CancelException()
    
    # protected
    def _pulse(self, new_message: str='') -> None:
        """
        Changes the progress bar in the visible progress dialog
        to be an indeterminate progress bar.
        
        Raises:
        * _CancelException
        """
        assert self._dialog is not None
        (ok, _) = self._dialog.Pulse(new_message)
        if not ok:
            raise self._CancelException()


# ------------------------------------------------------------------------------
# OpenProjectProgressListener

_active_progress_listener = None  # type: Optional[OpenProjectProgressListener]


class CancelOpenProject(Exception):
    pass


class VetoUpgradeProject(Exception):
    pass


# NOTE: See subclass OpenProjectProgressDialog for the documentation of
#       the various methods in this interface.
class OpenProjectProgressListener:
    def opening_project(self) -> None:
        pass
    
    def upgrading_project(self, message: str) -> None:
        pass
    
    def will_upgrade_revisions(self, approx_revision_count: int, can_veto: bool) -> None:
        pass
    
    def upgrading_revision(self, index: int, revisions_per_second: float) -> None:
        pass
    
    def did_upgrade_revisions(self, revision_count: int) -> None:
        pass
    
    def loading_root_resources(self, root_resource_count: int) -> None:
        pass
    
    def loading_resource_groups(self, resource_group_count: int) -> None:
        pass
    
    def loading_root_resource_views(self) -> None:
        pass
    
    def loading_resource_group_views(self) -> None:
        pass
    
    def creating_entity_tree_nodes(self, entity_tree_node_count: int) -> None:
        pass
    
    def reset(self) -> None:
        pass


DummyOpenProjectProgressListener = OpenProjectProgressListener


class OpenProjectProgressDialog(_AbstractProgressDialog, OpenProjectProgressListener):
    _dialog_title = 'Opening Project...'  # override
    _CancelException = CancelOpenProject  # override
    
    _approx_revision_count: float | None
    _resource_count: int | None
    _root_resource_count: int | None
    _resource_group_count: int | None
    _entity_tree_node_count: int | None
    
    # NOTE: Only changed when tests are running
    _always_show_upgrade_required_modal = False
    _upgrading_revision_progress = 0  # type: Optional[int]
    
    def __init__(self) -> None:
        super().__init__()
        self._approx_revision_count = None
        self._resource_count = None
        self._root_resource_count = None
        self._resource_group_count = None
        self._entity_tree_node_count = None
    
    # === Enter ===
    
    def __enter__(self) -> Self:
        return self
    
    # === Phase 0 ===
    
    @override
    def opening_project(self) -> None:
        self._show_noncancelable_indeterminate_message(
            f'Opening project...')
    
    def _show_noncancelable_indeterminate_message(self, initial_message: str) -> None:
        # HACK: wxGTK does not reliably update wx.ProgressDialog's message
        #       immediately after it is shown. So be sure to initialize
        #       the wx.ProgressDialog's message immediately to what it will
        #       be updated to soon.
        self._update_can_cancel(False, initial_message)
        
        # Change dialog to show an indeterminate progress bar
        try:
            self._pulse(initial_message)
        except CancelOpenProject:
            # Ignore cancel request
            pass
    
    # === Phase 1: Upgrade Revisions ===
    
    @override
    def upgrading_project(self, message: str) -> None:
        """
        Called immediately before a minor upgrade of a project is about to start.
        
        No progress will be reported during the minor upgrade.
        """
        self._show_noncancelable_indeterminate_message(
            f'Upgrading project: {message}')
    
    @override
    def will_upgrade_revisions(self, approx_revision_count: int, can_veto: bool) -> None:
        """
        Called immediately before a major upgrade of project revisions is about
        to start. If can_veto is True then the upgrade can be deferred by
        raising VetoUpgradeProject.
        
        If can_veto is False then VetoUpgradeProject must not be raised.
        
        Raises:
        * VetoUpgradeProject -- if user declines to upgrade the project
        * CancelOpenProject -- if user cancels opening the project
        """
        HISTORICAL_MIN_MIGRATION_SPEED = 200  # revisions/sec
        
        self._approx_revision_count = approx_revision_count
        
        initial_message = f'Upgrading about {approx_revision_count:n} revision(s)...'
        self._update_can_cancel(True, initial_message)
        
        assert self._dialog is not None
        self._dialog.SetRange(max(approx_revision_count, 1))
        self._update(0, initial_message)
        
        eta_total_minutes = approx_revision_count // HISTORICAL_MIN_MIGRATION_SPEED // 60
        if eta_total_minutes <= 2 and not self._always_show_upgrade_required_modal:
            # Automatically accept an upgrade if it looks like it will be fast
            pass
        else:
            # Prompt whether to start/continue upgrade now
            
            # TODO: Report ETA as "X hours Y minutes" rather than as
            #       just "Z minutes", since a several hours may be
            #       required for large projects
            #
            # TODO: Consider using self._dialog as the parent of this message
            #       dialog rather than making it its own top-level dialog
            dialog = wx.MessageDialog(None,
                message=(
                    f'Your project needs to be upgraded. '
                    f'About {eta_total_minutes:n} minutes will be required.'
                    # (TODO: Provide a 1 sentence summary of the benefits
                    #        of upgrading.)
                ),
                caption='Upgrade Required',
                style=(
                    wx.YES_NO|wx.CANCEL
                    if can_veto
                    else wx.OK|wx.CANCEL
                ),
            )
            dialog.Name = 'cr-upgrade-required'
            with dialog:
                if can_veto:
                    dialog.SetYesNoCancelLabels('Continue', '&Later', wx.ID_CLOSE)
                else:
                    dialog.SetOKCancelLabels('Continue', wx.ID_CLOSE)
                dialog.SetEscapeId(wx.ID_CANCEL)
                dialog.SetAcceleratorTable(wx.AcceleratorTable([
                    wx.AcceleratorEntry(wx.ACCEL_CTRL, ord('L'), wx.ID_NO),
                    wx.AcceleratorEntry(wx.ACCEL_CTRL, ord('W'), wx.ID_CANCEL),
                ]))
                position_dialog_initially(dialog)
                choice = ShowModal(dialog)
            if choice in [wx.ID_YES, wx.ID_OK]:
                pass
            elif choice == wx.ID_NO:
                raise VetoUpgradeProject()
            elif choice == wx.ID_CANCEL:
                raise CancelOpenProject()
            else:
                raise AssertionError()
        
        OpenProjectProgressDialog._upgrading_revision_progress = 0
    
    @override
    def upgrading_revision(self, index: int, revisions_per_second: float) -> None:
        """
        Called about once every second while project is being upgraded,
        to report progress of the upgrade.
        
        Raises:
        * CancelOpenProject
        """
        print(f'Upgrading revisions: {index:n} / {self._approx_revision_count:n} ({int(revisions_per_second):n} rev/sec)')
        self._update(index)
        OpenProjectProgressDialog._upgrading_revision_progress = index
    
    @override
    def did_upgrade_revisions(self, revision_count: int) -> None:
        """
        Called immediately after a major upgrade completes.
        """
        assert self._dialog is not None
        self._dialog.SetRange(max(revision_count, 1))
        OpenProjectProgressDialog._upgrading_revision_progress = None  # done
    
    # === Phase 2: Load ===
    
    # Step 1
    @override
    def loading_root_resources(self, root_resource_count: int) -> None:
        """
        Raises:
        * CancelOpenProject
        """
        self._root_resource_count = root_resource_count
        
        initial_message = f'Loading {root_resource_count:n} root resources(s)...'
        self._update_can_cancel(True, initial_message)
        
        assert self._dialog is not None
        self._dialog.SetRange(5)
        self._update(0, initial_message)
    
    # Step 2
    @override
    def loading_resource_groups(self, resource_group_count: int) -> None:
        """
        Raises:
        * CancelOpenProject
        """
        self._resource_group_count = resource_group_count
        
        assert self._dialog is not None
        self._update(
            1,
            f'Loading {resource_group_count:n} resource groups...')
    
    # Step 3
    @override
    def loading_root_resource_views(self) -> None:
        """
        Raises:
        * CancelOpenProject
        """
        if self._root_resource_count is None:
            return
        
        assert self._dialog is not None
        self._update(
            2,
            f'Creating {self._root_resource_count:n} root resource views...')
    
    # Step 4
    @override
    def loading_resource_group_views(self) -> None:
        """
        Raises:
        * CancelOpenProject
        """
        if self._resource_group_count is None:
            return
        
        assert self._dialog is not None
        self._update(
            3,
            f'Creating {self._resource_group_count} resource group views...')
    
    # Step 5
    @override
    def creating_entity_tree_nodes(self, entity_tree_node_count: int) -> None:
        """
        Raises:
        * CancelOpenProject
        """
        if self._root_resource_count is None or self._resource_group_count is None:
            return
        
        assert self._dialog is not None
        if entity_tree_node_count != (self._root_resource_count + self._resource_group_count):
            raise AssertionError('Unexpected number of initial entity tree nodes')
        self._entity_tree_node_count = entity_tree_node_count
        
        self._update(
            4,
            f'Creating {entity_tree_node_count} entity tree nodes...')
    
    # === Exit ===
    
    def __exit__(self, tp, value, tb) -> None:
        self.reset()


# ------------------------------------------------------------------------------
# LoadUrlsProgressListener

class CancelLoadUrls(Exception):
    pass


# NOTE: See subclass LoadUrlsProgressDialog for the documentation of
#       the various methods in this interface.
class LoadUrlsProgressListener:
    def will_load_resources(self, approx_resource_count: int) -> None:
        pass
    
    def loading_resource(self, index: int) -> None:
        pass
    
    def did_load_resources(self, resource_count: int) -> None:
        pass
    
    def indexing_resources(self) -> None:
        pass
    
    def reset(self) -> None:
        pass


DummyLoadUrlsProgressListener = LoadUrlsProgressListener


class LoadUrlsProgressDialog(_AbstractProgressDialog, LoadUrlsProgressListener):
    _dialog_title = 'Loading URLs...'  # override
    _CancelException = CancelLoadUrls  # override
    
    def __init__(self) -> None:
        super().__init__()
    
    @override
    def will_load_resources(self, approx_resource_count: int) -> None:
        """
        Called immediately before resources will be loaded.
        
        Raises:
        * CancelLoadUrls
        """
        initial_message = f'Loading about {approx_resource_count:n} resource(s)...'
        self._update_can_cancel(True, initial_message)
        
        assert self._dialog is not None
        self._dialog.SetRange(max(approx_resource_count, 1))
        self._update(0, initial_message)
    
    @override
    def loading_resource(self, index: int) -> None:
        """
        Called periodically while resources are being loaded, to report progress.
        
        Raises:
        * CancelLoadUrls
        """
        self._update(index)
    
    @override
    def did_load_resources(self, resource_count: int) -> None:
        """
        Called immediately after resources finished loading.
        """
        assert self._dialog is not None
        self._resource_count = resource_count
        self._dialog.SetRange(max(resource_count, 1))
    
    @override
    def indexing_resources(self) -> None:
        """
        Raises:
        * CancelLoadUrls
        """
        assert self._dialog is not None
        assert self._resource_count is not None
        message = f'Indexing {self._resource_count:n} resources(s)...'
        self._update(self._dialog.GetRange(), message)


# ------------------------------------------------------------------------------
# DeferredProgressDialog

_R = TypeVar('_R')

_Call = tuple[str, tuple[object, ...], dict[str, object]]

class DeferredProgressDialog:
    """
    Similar to wx.ProgressDialog but is hidden by default
    and auto-shows itself only after a delay.
    
    If the related operation completes quickly then no progress dialog
    will be shown at all.
    """
    def __init__(self,
            # wx.ProgressDialog parameters
            title,
            message,
            maximum=100,
            parent=None,
            style=wx.PD_APP_MODAL|wx.PD_AUTO_HIDE,
            
            # DeferredProgressDialog parameters
            show_after: float | None=None,
            
            # Unknown parameters
            *args,
            **kwargs) -> None:
        if show_after is None:
            raise ValueError('Expected show_after=... keyword argument')
        
        self._dialog = None  # type: Optional[wx.ProgressDialog]
        self._calls = [(
            '__init__',
            (title, message, maximum, parent, style) + args,
            kwargs
        )]  # type: List[_Call]
        self._show_at = time.time() + show_after
        
        self._shown = False
        self._name = None  # type: Optional[str]
        self._value = 0
        self._maximum = maximum
    
    @overload
    def _call(self, call: _Call) -> None: ...
    @overload
    def _call(self, call: _Call, default_result: _R) -> _R: ...
    def _call(self, call, default_result=None):
        if not self._shown:
            # Defer operation
            self._calls.append(call)
            return default_result
        else:
            # Apply operation
            return getattr(self._dialog, call[0])(*call[1], **call[2])
    
    # === Properties ===
    
    def _get_name(self) -> str | None:
        return self._name
    def _set_name(self, name: str) -> None:
        self._name = name
        self._call(('SetName', (name,), {}))
    Name = property(_get_name, _set_name)
    
    @property
    def Value(self) -> int:
        return self._value
    
    def SetRange(self, maximum: int) -> None:
        self._maximum = maximum
        self._call(('SetRange', (maximum,), {}))
    
    def GetRange(self) -> int:
        return self._maximum
    
    # === Operations ===
    
    def Show(self) -> None:
        if self._shown:
            return
        
        # Show now. Apply deferred operations.
        assert len(self._calls) >= 1
        assert self._calls[0][0] == '__init__'
        self._dialog = wx.ProgressDialog(*self._calls[0][1], **self._calls[0][2])
        self._shown = True
        for call in self._calls[1:]:
            self._call(call)
        self._calls.clear()
    
    def Destroy(self) -> None:
        # Defer/apply operation
        self._call(('Destroy', (), {}))
        
        if self._dialog is not None:
            self._dialog = None
        self._shown = False
    
    def Update(self, value: int, newmsg: str='') -> tuple[bool, bool]:
        # Show self if show_at time has passed
        if not self._shown and time.time() >= self._show_at:
            self.Show()
        
        # Defer/apply operation
        self._value = value
        return self._call(('Update', (value, newmsg), {}), (True, False))
    
    def Pulse(self, newmsg: str) -> tuple[bool, bool]:
        # Show self if show_at time has passed
        if not self._shown and time.time() >= self._show_at:
            self.Show()
        
        # Defer/apply operation
        return self._call(('Pulse', (newmsg,), {}), (True, False))


# ------------------------------------------------------------------------------



from collections.abc import Iterator
from contextlib import contextmanager
from unittest.mock import patch


@contextmanager
def database_on_ssd(is_ssd: bool) -> Iterator[None]:
    def mock_is_ssd(itempath: str) -> bool:
        return is_ssd
    
    with patch('crystal.util.ssd._is_mac_ssd', mock_is_ssd), \
            patch('crystal.util.ssd._is_linux_ssd', mock_is_ssd), \
            patch('crystal.util.ssd._is_windows_ssd', mock_is_ssd):
        yield
